/**
 * A bunch of simple, static functions to return DAQ conventions
 *
 * @author Jonathan Paley
 * @version $Id: DAQConventions.h,v 1.12 2013/02/07 20:20:53 jpaley Exp $
 */

#ifndef NOVADAQCONVENTIONS_H
#define NOVADAQCONVENTIONS_H

#include <string>
#include <iostream>
#include <fstream>
#include <stdint.h>

namespace novadaq
{
  namespace cnv {
    
    enum RunType {
      kTestRun,
      kCommissioningRun,
      kPedestalRun,
      kProductionRun,
      kNRunType
    };

    enum DetId {
      kUNKNOWN_DET,      
      kNEARDET,             //!< Near Detector in the NuMI cavern
      kFARDET,              //!< Far Detector at Ash River, MN
      kNDOS,                //!< Prototype Near Detector on the surface at FNAL
      kNDSBTEST,
      kTESTBEAM,
      kNDetector,
      kFCCDAQ
    };

    enum SubDetId {
      kUNKNOWN_SUBDET,
      k2Block_2x2,
      k3Block_2x2,
      k2Block_1x1,
      k3Block_1x1,
      kNSubDetector
    };

    class RunInfo {
    public:
      static std::string GetType(int rt) {
	switch (rt) {
	case kTestRun:
	  return std::string("Test");
	case kCommissioningRun:
	  return std::string("Commissioning");
	case kPedestalRun:
	  return std::string("Pedestal");
	case kProductionRun:
	  return std::string("Production");
	default:
	  break;
	}
	return std::string("");	
      }

      static int GetRunId(std::string rid) {
	if (rid == "Test") return kTestRun;
	if (rid == "Commissioning") return kCommissioningRun;
	if (rid == "Pedestal") return kPedestalRun;
	if (rid == "Production") return kProductionRun;
	return -1;
      }

      static bool RunIdOk(std::string rid) {
	return (RunInfo::GetRunId(rid) >= 0);
      }

      static bool RunTypeOk(int rt) {
	return (RunInfo::GetType(rt) != "");
      }

    };

    class DetInfo {
    public:
      static std::string GetName(int id) { 
	switch (id) {  
	case kNEARDET:
	  return std::string("NearDet");
	case kFARDET:
	  return std::string("FarDet");
	case kNDOS:
	  return std::string("NDOS");
	case kNDSBTEST:
	  return std::string("NDSBTest");
	case kTESTBEAM:
	  return std::string("TestBeam");
	case kFCCDAQ:
	  return std::string("FCCDAQ");
	case kUNKNOWN_DET:
	default:
	  break;
	}
	return std::string("Unknown");
      }
      
      static int GetId(std::string name) {
	if (name == "NearDet") return kNEARDET;
	if (name == "FarDet") return kFARDET;
	if (name == "NDOS") return kNDOS;
	if (name == "NDSBTest") return kNDSBTEST;
	if (name == "TestBeam") return kTESTBEAM;
	if (name == "FCCDAQ") return kFCCDAQ;
	return kUNKNOWN_DET;
      }

      static bool DetIdOk(int id) {
	return (DetInfo::GetName(id) != "Unknown");
      }

      static bool DetNameOk(std::string name) {
	return (DetInfo::GetId(name) != kUNKNOWN_DET);
      }
    };

    class SubDetInfo {
    public:
      static std::string GetName(int id) { 
	switch (id) {  
	case k2Block_2x2:
	  return std::string("_2Block_2x2");
	case k3Block_2x2:
	  return std::string("_3Block_2x2");
	case k2Block_1x1:
	  return std::string("_2Block_1x1");
	case k3Block_1x1:
	  return std::string("_3Block_1x1");
	case kUNKNOWN_SUBDET:
	default:
	  break;
	}
	return std::string("Unknown");
      }
      
      static int GetId(std::string name) {
	if (name == "_2Block_2x2") return k2Block_2x2;
	if (name == "_3Block_2x2") return k3Block_2x2;
	if (name == "_2Block_1x1") return k2Block_1x1;
	if (name == "_3Block_1x1") return k3Block_1x1;
	return kUNKNOWN_SUBDET;
      }

    };

  }
}

#endif // NOVADAQCONVENTIONS_H
