#include <DatabaseUtils/DCS/APDSettings.h>
#include <DatabaseUtils/DAQConfig/DBColumnParams.h>
#include <DatabaseUtils/DAQConfig/NamedConfigUtils.h>
#include <DatabaseUtils/DAQConfig/SubsystemNameSpec.h>
#include <algorithm>

using namespace dbutils::daqconfig;

namespace dbutils {
namespace dcs {

const std::string APDSettings::WARM_COOLING_STATE("warm");
const std::string APDSettings::COLD_COOLING_STATE("cold");

/**
 * Constructs an APDSettings instance for the specified global
 * configuration ID and list of DCMs.
 */
APDSettings::
APDSettings(const int64_t& globalConfigId,
            const std::vector<std::string>& dcmList)
{
    _hasValidData = false;

    IDSpec idSpec(globalConfigId, IDSpec::GLOBAL, IDSpec::SLOWCONTROLS);
    std::vector<int64_t> subsysConfigIdList;
    bool status = NamedConfigUtils::
        getSubsystemConfigIdAncestryFromGlobalId(idSpec,
                                                 "APDHardware",
                                                 "slowcontrols",
                                                 subsysConfigIdList);
    if (! status) {return;}

    _dcmList = dcmList;
    std::sort(_dcmList.begin(), _dcmList.end());

    _hasValidData = _fetchAllData(subsysConfigIdList);
}

/**
 * Constructs an APDSettings instance for the specified global
 * configuration name specification and list of DCMs.
 */
APDSettings::
APDSettings(GlobalNameSpec nameSpec,
            const std::vector<std::string>& dcmList)
{
    _hasValidData = false;

    // just in case...
    nameSpec.isSlowControlsConfiguration(true);

    std::vector<int64_t> subsysConfigIdList;
    int64_t subsystemConfigId;
    bool status = NamedConfigUtils::
        getSubsystemConfigIdFromGlobalName(nameSpec, "APDHardware",
                                           "slowcontrols", subsystemConfigId);
    if (status) {
        IDSpec idSpec(subsystemConfigId, IDSpec::SUBSYSTEM,
                      IDSpec::SLOWCONTROLS);
        status = NamedConfigUtils::
            getConfigIdAncestry(idSpec, subsysConfigIdList);
    }
    if (! status) {return;}

    _dcmList = dcmList;
    std::sort(_dcmList.begin(), _dcmList.end());

    _hasValidData = _fetchAllData(subsysConfigIdList);
}

/**
 * Constructs an APDSettings instance for the specified subsystem
 * configuration name specification and list of DCMs.
 */
APDSettings::
APDSettings(SubsystemNameSpec nameSpec,
            const std::vector<std::string>& dcmList)
{
    _hasValidData = false;

    // just in case...
    nameSpec.isSlowControlsConfiguration(true);

    std::vector<int64_t> subsysConfigIdList;
    bool status = NamedConfigUtils::
        getSubsystemConfigIdAncestry(nameSpec, subsysConfigIdList);
    if (! status) {return;}

    _dcmList = dcmList;
    std::sort(_dcmList.begin(), _dcmList.end());

    _hasValidData = _fetchAllData(subsysConfigIdList);
}

/**
 * Fetches the temperature setpoint (as an int) for the 
 * specified DCM name and FEB ID.
 *
 * @return true if the value was successfully determined, false if not.
 */
bool APDSettings::
getTemperature(const std::string& dcmName, const int& febId,
               int& temperature)
{
    if (_temperatureDataSet.get() == 0) {return false;}
    if (! _temperatureDataSet->hasValidData()) {return false;}

    double doubleTemp;
    bool status =
        _temperatureDataSet->getEffectiveValueFor2Keys(dcmName, febId,
                                                       "temperature",
                                                       doubleTemp);
    if (! status) {return false;}

    temperature = (int) doubleTemp;
    return true;
}

/**
 * Fetches whether cooling is enabled for the specified DCM name and FEB ID.
 *
 * @return true if the value was successfully determined, false if not.
 */
bool APDSettings::
getCoolingEnableFlag(const std::string& dcmName, const int& febId,
                     bool& isEnabled)
{
    if (_enableMaskDataSet.get() == 0) {return false;}
    if (! _enableMaskDataSet->hasValidData()) {return false;}

    int64_t fullMask;
    if (! _enableMaskDataSet->getEffectiveValueFor1Key(dcmName, "mask",
                                                       fullMask)) {
        return false;
    }

    int64_t testMask = (((int64_t)1) << febId);
    isEnabled = (febId >= 0 && febId < 64 && (fullMask & testMask) != 0);
    return true;
}

/**
 * Fetches the voltage regulator setting for the specified DCM name and
 * FEB ID, taking into account whether cooling is enabled for the FEB.
 *
 * @return true if the value was successfully determined, false if not.
 */
bool APDSettings::
getVoltage(const std::string& dcmName, const int& febId,
           int& regulatorSetting)
{
    if (_hvAdjustDataSet.get() == 0) {return false;}
    if (! _hvAdjustDataSet->hasValidData()) {return false;}

    std::string coolingState = WARM_COOLING_STATE;
    bool isEnabled;
    if (getCoolingEnableFlag(dcmName, febId, isEnabled)) {
        if (isEnabled) {
            coolingState = COLD_COOLING_STATE;
        }
    }

    return _hvAdjustDataSet->getEffectiveValueFor3Keys(coolingState,
                                                       dcmName, febId,
                                                       "voltage",
                                                       regulatorSetting);
}

/**
 * Fetches the temperature setpoint (as a double) for the 
 * specified DCM name and FEB ID.
 *
 * @return true if the value was successfully determined, false if not.
 */
bool APDSettings::
getTemperature(const std::string& dcmName, const int& febId,
               double& temperature)
{
    if (_temperatureDataSet.get() == 0) {return false;}
    if (! _temperatureDataSet->hasValidData()) {return false;}
    return _temperatureDataSet->getEffectiveValueFor2Keys(dcmName, febId,
                                                          "temperature",
                                                          temperature);
}

/**
 * Fetches whether cooling is enabled for the specified DCM name and FEB ID.
 *
 * @return true if the value was successfully determined, false if not.
 */
bool APDSettings::
getTemperatureEnable(const std::string& dcmName, const int& febId,
                     bool& enableFlag)
{
    if (_enableMaskDataSet.get() == 0) {return false;}
    if (! _enableMaskDataSet->hasValidData()) {return false;}

    int64_t fullMask;
    if (! _enableMaskDataSet->getEffectiveValueFor1Key(dcmName, "mask",
                                                       fullMask)) {
        return false;
    }

    int64_t testMask = (((int64_t)1) << febId);
    enableFlag = (febId >= 0 && febId < 64 && (fullMask & testMask) != 0);
    return true;
}

/**
 * Fetches the voltage regulator setting for the specified DCM name, 
 * FEB ID, and cooling state.
 *
 * @return true if the value was successfully determined, false if not.
 */
bool APDSettings::
getVoltage(const std::string& dcmName, const int& febId,
           const std::string& coolingState, int& regulatorSetting)
{
    if (_hvAdjustDataSet.get() == 0) {return false;}
    if (! _hvAdjustDataSet->hasValidData()) {return false;}
    return _hvAdjustDataSet->getEffectiveValueFor3Keys(coolingState,
                                                       dcmName, febId,
                                                       "voltage",
                                                       regulatorSetting);
}

/**
 * Fetches the list of temperature setpoints (as integers) for the 
 * FEBs that are connected to the specified DCM.  The resulting list will
 * be populated with 64 values, one for each FEB (in numerical order),
 * when this method executes successfully.
 *
 * @return true if the requested values were successfully determined,
 *         false if not.
 */
bool APDSettings::
getTemperatures(const std::string& dcmName,
                std::vector<int>& temperatureList)
{
    if (_temperatureDataSet.get() == 0) {return false;}
    if (! _temperatureDataSet->hasValidData()) {return false;}

    temperatureList.clear();
    for (uint16_t febId = DBColumnParams::LOW_FEB_ID;
         febId <= DBColumnParams::HIGH_FEB_ID; ++febId) {
        double tmpValue;
        if (! _temperatureDataSet->getEffectiveValueFor2Keys(dcmName, febId,
                                                             "temperature",
                                                             tmpValue)) {
            return false;
        }
        temperatureList.push_back((int) tmpValue);
    }
    return true;
}

/**
 * Fetches the cooling enable mask for the FEBs that are connected
 * to the specified DCM.  The resulting mask contains the information
 * for all 64 FEBs, in numerical order, with the least significant bit
 * in the mask corresponding to FEB 0.
 *
 * @return true if the requested mask was successfully determined,
 *         false if not.
 */
bool APDSettings::
getCoolingEnableMask(const std::string& dcmName, int64_t& enableMask)
{
    if (_enableMaskDataSet.get() == 0) {return false;}
    if (! _enableMaskDataSet->hasValidData()) {return false;}
    return _enableMaskDataSet->getEffectiveValueFor1Key(dcmName, "mask",
                                                        enableMask);
}

/**
 * Fetches the list of voltage regulator settings for the FEBs that are
 * connected to the specified DCM, taking into account whether cooling
 * is enabled for each FEB.  The resulting list will
 * be populated with 64 values, one for each FEB (in numerical order),
 * when this method executes successfully.
 *
 * @return true if the requested values were successfully determined,
 *         false if not.
 */
bool APDSettings::
getVoltages(const std::string& dcmName, std::vector<int>& voltageList)
{
    if (_hvAdjustDataSet.get() == 0) {return false;}
    if (! _hvAdjustDataSet->hasValidData()) {return false;}

    int64_t coolingEnableMask;
    if (! getCoolingEnableMask(dcmName, coolingEnableMask)) {
        return false;
    }
    int64_t _64BitOne = 1;

    voltageList.clear();
    for (uint16_t febId = DBColumnParams::LOW_FEB_ID;
         febId <= DBColumnParams::HIGH_FEB_ID; ++febId) {

        std::string coolingState = WARM_COOLING_STATE;
        if ((coolingEnableMask & (_64BitOne << febId)) != 0) {
            coolingState = COLD_COOLING_STATE;
        }

        int tmpValue;
        if (! _hvAdjustDataSet->getEffectiveValueFor3Keys(coolingState,
                                                          dcmName, febId,
                                                          "voltage",
                                                          tmpValue)) {
            return false;
        }
        voltageList.push_back(tmpValue);
    }
    return true;
}

/**
 * Fetches the list of temperature setpoints (as doubles) for the 
 * FEBs that are connected to the specified DCM.  The resulting list will
 * be populated with 64 values, one for each FEB (in numerical order),
 * when this method executes successfully.
 *
 * @return true if the requested values were successfully determined,
 *         false if not.
 */
bool APDSettings::
getTemperatures(const std::string& dcmName,
                std::vector<double>& temperatureList)
{
    if (_temperatureDataSet.get() == 0) {return false;}
    if (! _temperatureDataSet->hasValidData()) {return false;}

    temperatureList.clear();
    for (uint16_t febId = DBColumnParams::LOW_FEB_ID;
         febId <= DBColumnParams::HIGH_FEB_ID; ++febId) {
        double tmpValue;
        if (! _temperatureDataSet->getEffectiveValueFor2Keys(dcmName, febId,
                                                             "temperature",
                                                             tmpValue)) {
            return false;
        }
        temperatureList.push_back(tmpValue);
    }
    return true;
}

/**
 * Fetches the cooling enable mask for the FEBs that are connected
 * to the specified DCM.  The resulting mask contains the information
 * for all 64 FEBs, in numerical order, with the least significant bit
 * in the mask corresponding to FEB 0.
 *
 * @return true if the requested mask was successfully determined,
 *         false if not.
 */
bool APDSettings::
getTemperatureEnableMask(const std::string& dcmName, int64_t& enableMask)
{
    if (_enableMaskDataSet.get() == 0) {return false;}
    if (! _enableMaskDataSet->hasValidData()) {return false;}
    return _enableMaskDataSet->getEffectiveValueFor1Key(dcmName, "mask",
                                                        enableMask);
}

/**
 * Fetches the list of warm or cold voltage regulator settings for the 
 * FEBs that are connected to the specified DCM.  The resulting list will
 * be populated with 64 values, one for each FEB (in numerical order),
 * when this method executes successfully.
 *
 * @return true if the requested values were successfully determined,
 *         false if not.
 */
bool APDSettings::
getVoltages(const std::string& dcmName, const std::string& coolingState,
            std::vector<int>& voltageList)
{
    if (_hvAdjustDataSet.get() == 0) {return false;}
    if (! _hvAdjustDataSet->hasValidData()) {return false;}

    voltageList.clear();
    for (uint16_t febId = DBColumnParams::LOW_FEB_ID;
         febId <= DBColumnParams::HIGH_FEB_ID; ++febId) {
        int tmpValue;
        if (! _hvAdjustDataSet->getEffectiveValueFor3Keys(coolingState,
                                                          dcmName, febId,
                                                          "voltage",
                                                          tmpValue)) {
            return false;
        }
        voltageList.push_back(tmpValue);
    }
    return true;
}

/**
 * Fetches the dry air channel for the specified DCM name and FEB ID.
 *
 * @return true if the value was successfully determined, false if not.
 */
bool APDSettings::
getDryAirChannel(const std::string& dcmName, const int& febId,
                 int& channelNumber)
{
    if (_dryAirChannelDataSet.get() == 0) {return false;}
    if (! _dryAirChannelDataSet->hasValidData()) {return false;}
    return _dryAirChannelDataSet->getEffectiveValueFor2Keys(dcmName, febId,
                                                            "dryairchannel",
                                                            channelNumber);
}

/**
 * Fetches the list of dry air channels for the 
 * FEBs that are connected to the specified DCM.  The resulting list will
 * be populated with 64 values, one for each FEB (in numerical order),
 * when this method executes successfully.
 *
 * @return true if the requested values were successfully determined,
 *         false if not.
 */
bool APDSettings::
getDryAirChannels(const std::string& dcmName, std::vector<int>& channelList)
{
    if (_dryAirChannelDataSet.get() == 0) {return false;}
    if (! _dryAirChannelDataSet->hasValidData()) {return false;}

    channelList.clear();
    for (uint16_t febId = DBColumnParams::LOW_FEB_ID;
         febId <= DBColumnParams::HIGH_FEB_ID; ++febId) {
        int tmpValue;
        if (! _dryAirChannelDataSet->getEffectiveValueFor2Keys(dcmName, febId,
                                                               "dryairchannel",
                                                               tmpValue)) {
            return false;
        }
        channelList.push_back(tmpValue);
    }
    return true;
}

/**
 * Fetches the alarm threshold for the specified DCM name, FEB ID,
 * PV name, and threshold type (hihi, hilo, lohi, or lolo).
 *
 * @return true if the value was successfully determined, false if not.
 */
bool APDSettings::
getAlarmThreshold(const std::string& dcmName,
                  const int& febId,
                  const std::string& pvName,
                  const std::string& thresholdType,
                  double& threshold)
{
    if (_alarmThresholdDataSet.get() == 0) {return false;}
    if (! _alarmThresholdDataSet->hasValidData()) {return false;}
    return _alarmThresholdDataSet->
        getEffectiveValueFor3Keys(pvName, dcmName, febId,
                                  thresholdType, threshold);
}

bool APDSettings::
_fetchAllData(const std::vector<int64_t>& subsysConfigIdList)
{
    {
        ConfigDataSelector cfgSelector;
        cfgSelector.addSelectionColumn(DBColumnParams::DCMID_COLUMN_NAME,
                                       _dcmList[0],_dcmList[_dcmList.size()-1]);
        cfgSelector.addSelectionColumn(DBColumnParams::FEBID_COLUMN_NAME);

        std::vector<std::string> dataColumnsOfInterest;
        dataColumnsOfInterest.push_back("temperature");

        std::string fname("NovaDatabase/tables/DCS/APDTemperatureSettings.xml");
        _temperatureDataSet.
            reset(new ConfigDataTree(fname, subsysConfigIdList, cfgSelector,
                                     dataColumnsOfInterest));
    }

    {
        ConfigDataSelector cfgSelector;
        cfgSelector.addSelectionColumn(DBColumnParams::DCMID_COLUMN_NAME,
                                       _dcmList[0],_dcmList[_dcmList.size()-1]);

        std::vector<std::string> dataColumnsOfInterest;
        dataColumnsOfInterest.push_back("mask");

        std::string fname("NovaDatabase/tables/DCS/DCSFEBEnableMasks.xml");
        _enableMaskDataSet.
            reset(new ConfigDataTree(fname, subsysConfigIdList, cfgSelector,
                                     dataColumnsOfInterest));
    }

    {
        ConfigDataSelector cfgSelector;
        cfgSelector.addSelectionColumn("coolingstate");
        cfgSelector.addSelectionColumn(DBColumnParams::DCMID_COLUMN_NAME,
                                       _dcmList[0],_dcmList[_dcmList.size()-1]);
        cfgSelector.addSelectionColumn(DBColumnParams::FEBID_COLUMN_NAME);

        std::vector<std::string> dataColumnsOfInterest;
        dataColumnsOfInterest.push_back("voltage");

        std::string fname("NovaDatabase/tables/DCS/APDHighVoltages.xml");
        _hvAdjustDataSet.
            reset(new ConfigDataTree(fname, subsysConfigIdList, cfgSelector,
                                     dataColumnsOfInterest));
    }

    {
        ConfigDataSelector cfgSelector;
        cfgSelector.addSelectionColumn(DBColumnParams::DCMID_COLUMN_NAME,
                                       _dcmList[0],_dcmList[_dcmList.size()-1]);
        cfgSelector.addSelectionColumn(DBColumnParams::FEBID_COLUMN_NAME);

        std::vector<std::string> dataColumnsOfInterest;
        dataColumnsOfInterest.push_back("dryairchannel");

        std::string fname("NovaDatabase/tables/DCS/FEBDryAirChannels.xml");
        _dryAirChannelDataSet.
            reset(new ConfigDataTree(fname, subsysConfigIdList, cfgSelector,
                                     dataColumnsOfInterest));
    }

    {
        ConfigDataSelector cfgSelector;
        cfgSelector.addSelectionColumn("pvName");
        cfgSelector.addSelectionColumn(DBColumnParams::DCMID_COLUMN_NAME,
                                       _dcmList[0],_dcmList[_dcmList.size()-1]);
        cfgSelector.addSelectionColumn(DBColumnParams::FEBID_COLUMN_NAME);

        std::vector<std::string> dataColumnsOfInterest;
        dataColumnsOfInterest.push_back("hihi");
        dataColumnsOfInterest.push_back("hilo");
        dataColumnsOfInterest.push_back("lohi");
        dataColumnsOfInterest.push_back("lolo");

        std::string fname("NovaDatabase/tables/DCS/EPICSAlarmThresholds.xml");
        _alarmThresholdDataSet.
            reset(new ConfigDataTree(fname, subsysConfigIdList, cfgSelector,
                                     dataColumnsOfInterest));
    }

    return _temperatureDataSet.get() != 0 && _temperatureDataSet->hasValidData() &&
        _enableMaskDataSet.get() != 0 && _enableMaskDataSet->hasValidData() &&
        _hvAdjustDataSet.get() != 0 && _hvAdjustDataSet->hasValidData() &&
        _dryAirChannelDataSet.get() != 0 && _dryAirChannelDataSet->hasValidData() &&
        _alarmThresholdDataSet.get() != 0 && _alarmThresholdDataSet->hasValidData();
}

};  // end of namespace dcs
};  // end of namespace dbutils
