#include <iostream>
#include <fstream>
#include <sys/stat.h>
#include <cstdlib>
#include <cctype>
#include <ctime>
#include <algorithm>

#include <NovaDatabase/Column.h>
#include <NovaDatabase/Util.h>

using namespace nova::database;
using namespace std;

//************************************************************

Column::Column(const dBColumn_t &c) :
  fIsNull(true), fModified(false), fTolType(kUnknown), fTolerance(0.)
{
  fValue = "";
  fName = c.getName();
  fType = "string";
  fCanBeNull = true;
  if (c.hasNull()) {
    fCanBeNull = c.getNull();
  }
  if (c.hasColType()) 
    fType = c.getColType();
  if (c.hasValue()) {
    fIsNull = false;
    fValue = c.getValue();
  }
  if (c.hasTolerance()) {
    fTolerance = c.getTolerance().getTolValue();
    std::string tolStr = c.getTolerance().getTolType();
    for (unsigned int i=0; i<tolStr.length(); ++i)
      tolStr[i] = tolower(tolStr[i]);
    if (tolStr == "fractional")
      fTolType = kFractional;
    else if (tolStr == "absolute")
      fTolType = kAbsolute;
    else 
      fTolType = kUnknown;
  }
}

//************************************************************

Column::~Column()
{

}

//************************************************************
void Column::Clear() 
{
  fValue = "";
  fIsNull = true;
  fModified = false; 
}

//************************************************************
bool Column::operator >= (const Column& c) const
{
  if (c.Type() != fType) return false;

  std::string val1 = fValue;
  std::string val2 = c.fValue;
  
  if (fType == "bool") {
    bool a = (val1 == "TRUE");
    bool b = (val2 == "TRUE");
    return ((a == b) || (a&!b));
  }

  if (fType == "int" || fType == "bigint" || fType == "long" ||
      fType == "short" || fType == "ulong") {
    long a = boost::lexical_cast<long>(val1);
    long b = boost::lexical_cast<long>(val2);
    return (a >= b);
  }

  if (fType == "float" || fType == "double") {
    double a = boost::lexical_cast<double>(val1);
    double b = boost::lexical_cast<double>(val2);
    if (fTolType == kAbsolute) 
      return ((a-b) >= fTolerance);
    else if (fTolType == kFractional)
      return (((a-b)/a) >= fTolerance);
    else
      return (a >= b);
  }

  if (fType == "string") 
    return (val1 >= val2);

  if (fType == "timestamp") {
    string a, b;
    if (this->Get(a) && c.Get(b)) {
      time_t tta, ttb;
      if (Util::TimeAsStringToTime_t(a,tta) && 
	  Util::TimeAsStringToTime_t(b,ttb) ) 
	return (tta >= ttb);
    }
    else 
      return false;
  }

  if (fType == "date") {
    string a, b;
    if (this->Get(a) && c.Get(b)) {
      time_t tta, ttb;
      if (Util::DateAsStringToTime_t(a,tta) && 
	  Util::DateAsStringToTime_t(b,ttb) ) 
	return (tta >= ttb);
    }
    else 
      return false;
  }
  return false;

}

//************************************************************
bool Column::operator > (const Column& c) const
{
  if (c.Type() != fType) return false;

  std::string val1 = fValue;
  std::string val2 = c.fValue;
  
  if (fType == "bool") {
    bool a = (val1 == "TRUE");
    bool b = (val2 == "TRUE");
    return (a&!b);
  }

  if (fType == "int" || fType == "bigint" || fType == "long" ||
      fType == "short" || fType == "ulong") {
    long a = boost::lexical_cast<long>(val1);
    long b = boost::lexical_cast<long>(val2);
    return (a > b);
  }

  if (fType == "float" || fType == "double") {
    double a = boost::lexical_cast<double>(val1);
    double b = boost::lexical_cast<double>(val2);
    if (fTolType == kAbsolute) 
      return ((a-b) > fTolerance);
    else if (fTolType == kFractional)
      return (((a-b)/a) > fTolerance);
    else
      return (a > b);
  }

  if (fType == "string") 
    return (val1 > val2);

  if (fType == "timestamp") {
    string a, b;
    if (this->Get(a) && c.Get(b)) {
      time_t tta, ttb;
      if (Util::TimeAsStringToTime_t(a,tta) && 
	  Util::TimeAsStringToTime_t(b,ttb) ) 
	return (tta > ttb);
    }
    else 
      return false;
  }

  if (fType == "date") {
    string a, b;
    if (this->Get(a) && c.Get(b)) {
      time_t tta, ttb;
      if (Util::DateAsStringToTime_t(a,tta) && 
	  Util::DateAsStringToTime_t(b,ttb) ) 
	return (tta > ttb);
    }
    else 
      return false;
  }
  return false;

}

//************************************************************
bool Column::operator <= (const Column& c) const
{
  if (c.fType != fType) return false;

  std::string val1 = fValue;
  std::string val2 = c.fValue;
  
  if (fType == "bool") {
    bool a = (val1 == "TRUE");
    bool b = (val2 == "TRUE");
    return ((a == b) || (!a&b));
  }

  if (fType == "int" || fType == "bigint" || fType == "long" ||
      fType == "short" ||  fType == "ulong") {
    long a = boost::lexical_cast<long>(val1);
    long b = boost::lexical_cast<long>(val2);
    return (a <= b);
  }

  if (fType == "float" || fType == "double") {
    double a = boost::lexical_cast<double>(val1);
    double b = boost::lexical_cast<double>(val2);
    if (fTolType == kAbsolute) 
      return ((a-b) <= fTolerance);
    else if (fTolType == kFractional)
      return (((a-b)/a) <= fTolerance);
    else
      return (a <= b);
  }

  if (fType == "string") 
    return (val1 <= val2);

  if (fType == "timestamp") {
    string a, b;
    if (this->Get(a) && c.Get(b)) {
      time_t tta, ttb;
      if (Util::TimeAsStringToTime_t(a,tta) && 
	  Util::TimeAsStringToTime_t(b,ttb) ) 
	return (tta <= ttb);
    }
    else 
      return false;
  }

  if (fType == "date") {
    string a, b;
    if (this->Get(a) && c.Get(b)) {
      time_t tta, ttb;
      if (Util::DateAsStringToTime_t(a,tta) && 
	  Util::DateAsStringToTime_t(b,ttb) ) 
	return (tta <= ttb);
    }
    else 
      return false;
  }

  return false;
  
}
//************************************************************
bool Column::operator < (const Column& c) const
{
  if (c.fType != fType) return false;

  std::string val1 = fValue;
  std::string val2 = c.fValue;
  
  if (fType == "bool") {
    bool a = (val1 == "TRUE");
    bool b = (val2 == "TRUE");
    return (!a&b);
  }

  if (fType == "int" || fType == "bigint" || fType == "long" ||
      fType == "short" ||  fType == "ulong") {
    long a = boost::lexical_cast<long>(val1);
    long b = boost::lexical_cast<long>(val2);
    return (a < b);
  }

  if (fType == "float" || fType == "double") {
    double a = boost::lexical_cast<double>(val1);
    double b = boost::lexical_cast<double>(val2);
    if (fTolType == kAbsolute) 
      return ((a-b) < fTolerance);
    else if (fTolType == kFractional)
      return (((a-b)/a) < fTolerance);
    else
      return (a < b);
  }

  if (fType == "string") 
    return (val1 < val2);

  if (fType == "timestamp") {
    string a, b;
    if (this->Get(a) && c.Get(b)) {
      time_t tta, ttb;
      if (Util::TimeAsStringToTime_t(a,tta) && 
	  Util::TimeAsStringToTime_t(b,ttb) ) 
	return (tta < ttb);
    }
    else 
      return false;
  }

  if (fType == "date") {
    string a, b;
    if (this->Get(a) && c.Get(b)) {
      time_t tta, ttb;
      if (Util::DateAsStringToTime_t(a,tta) && 
	  Util::DateAsStringToTime_t(b,ttb) ) 
	return (tta < ttb);
    }
    else 
      return false;
  }

  return false;
  
}
//************************************************************
bool Column::operator == (const Column& c) const
{
  if (c.fType != fType) return false;

  bool retVal = (fValue==c.fValue);

  if (fType == "float" || fType == "double") {
    if (fTolerance > 0.) {
      double a = boost::lexical_cast<double>(fValue);
      double b = boost::lexical_cast<double>(c.fValue);
      if (a == 0.)
	retVal = (a == b);
      else {
	if (fTolType == kAbsolute) 
	  retVal = (fabs(a-b) < fTolerance);
	else if (fTolType == kFractional)
	  retVal = (fabs((a-b)/a) < fTolerance);
      }
    }
  }
  
  return retVal;

}
