#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <sys/stat.h>
#include <cstdlib>
#include <cctype>
#include <algorithm>
#include <ctime>
#include <curl/curl.h>

#include <boost/tokenizer.hpp>
#include <boost/algorithm/string.hpp>
#include <boost/algorithm/string/case_conv.hpp>

#include <NovaDAQConventions/DAQConventions.h>
#include <NovaDAQUtilities/LocationUtils.h>
#include <NovaDatabase/Table.h>
#include <NovaDatabase/Util.h>

using namespace nova::database;
using namespace std;

boost::mutex Table::_xsdLock;

//************************************************************

Table::Table() : fMaxConnAttempts(10), fDetector("")
{  
  fDBHost="";
  fDBName="";
  fDBPort="";
  fUser="";
  fSchema="undef";
  fTableType = kGenericTable;

  fIgnoreEnvVar = false;
  fHasValidXML = false;
  fTestedXML = false;
  fTestedExists = false;
  fExistsInDB = false;
  addInsertTime = addUpdateTime = false;
  addInsertUser = addUpdateUser = false;

  Reset();

  char* tmpStr = getenv("NOVADBNCNXTRIES");
  if (tmpStr) {
    int nTry = atoi(tmpStr);
    if (nTry > 0)
      fMaxConnAttempts = nTry;
  }
  
}

//************************************************************

Table::Table(string tname) : fMaxConnAttempts(10), fDetector("")
{  
  fDBHost="";
  fDBName="";
  fDBPort="";
  fUser="";
  fSchema="undef";
  fTableType = kGenericTable;
  fIgnoreEnvVar = false;
  fHasValidXML = false;
  fTestedXML = false;
  fTestedExists = false;
  fExistsInDB = false;
  Reset();
  CreateFromXML(tname.c_str());
}

//************************************************************

Table::Table(const char* tname) : fMaxConnAttempts(10), fDetector("")
{
  fDBHost="";
  fDBName="";
  fDBPort="";
  fUser="";
  fSchema="undef";
  fTableType = kGenericTable;
  fIgnoreEnvVar = false;
  fHasValidXML = false;
  fTestedXML = false;
  fTestedExists = false;
  fExistsInDB = false;
  Reset();
  CreateFromXML(tname);
}

//************************************************************

Table::~Table()
{
  this->Clear();
  if (fHasConnection) CloseConnection();
}

//************************************************************

bool Table::CreateFromDB(const char* schemaName, const char* tableName, 
			 const char* dbhost, const char* dbname,    
			 const char* dbport, const char* dbuser)
{
  std::string errStr;

  if (!tableName) {
    errStr = "Table::CreateFromDB: missing table name!";
    throw std::runtime_error(errStr);
  }

  if (!schemaName) {
    errStr = "Table::CreateFromDB: missing table schema (eg, detector name)!";
    throw std::runtime_error(errStr);
  }

  if (dbname) SetDBName(dbname);
  if (DBName() == "") {
    errStr = "Table::CreateFromDB: missing database name!";
    throw std::runtime_error(errStr);
  }

  if (dbhost) SetDBHost(dbhost);
  if (DBHost() == "") {
    errStr = "Table::CreateFromDB: missing database host!";
    throw std::runtime_error(errStr);
  }

  if (dbport) SetDBPort(dbport);
  if (dbuser) SetUser(dbuser);

  fTable.reset(new dBTable());
  fTable->setTableName(tableName);
  fSchema = std::string(schemaName);
  boost::to_lower(fSchema);

  std::string stName = fSchema + std::string(".") + std::string(tableName);
  fIgnoreEnvVar = true;

  if (!ExistsInDB()) {
    errStr = "Table::CreateFromDB: table \'" + stName + "\' not found in database!";
    throw std::runtime_error(errStr);
  }

  Reset();
  fCol.clear();
  fXMLOverDB = false;

  bool hasConn = fHasConnection;
  if (! fHasConnection) {
    GetConnection();
    hasConn = false;
  }    

  std::vector<std::string> pkeyList;
  dBTable::dBPKey_type pkey;
  // get list of rows that are primary keys
  std::string cmd = "SELECT pg_attribute.attname, format_type(pg_attribute.atttypid, pg_attribute.atttypmod) FROM pg_index, pg_class, pg_attribute WHERE indrelid = pg_class.oid AND pg_attribute.attrelid = pg_class.oid AND pg_attribute.attnum = any(pg_index.indkey) AND indisprimary AND pg_class.oid = '" + stName + "'::regclass";

  PGresult* res = PQexec(fConnection,cmd.c_str());
  
  if (PQresultStatus(res) != PGRES_TUPLES_OK) {
    if (fVerbosity > 0)
      std::cerr << "Table::CreateFromDB command failed: " 
	   << PQerrorMessage(fConnection) << std::endl;
    PQclear(res);
    fExistsInDB = false;  // set this to false just in case
    CloseConnection();
    return false;
  }

  int nRow = PQntuples(res);
  if (nRow == 0) {
    std::cerr << "Table::CreateFromDB: no primary keys defined for table \'" << tableName << "\', unable to proceed." << std::endl;
    PQclear(res);
    fExistsInDB = false;
    CloseConnection();
    return false;
  }

  for (int i=0; i<nRow; ++i) {
    std::string key = string(PQgetvalue(res,i,0));
    pkeyList.push_back(key);
  }

  PQclear(res);

  // now get names and types of all columns
  cmd = "SELECT column_name, data_type from information_schema.columns where table_name = \'" + std::string(tableName) + "\' and table_schema=\'" + fSchema + "\'";
  res = PQexec(fConnection,cmd.c_str());

  if (PQresultStatus(res) != PGRES_TUPLES_OK) {
    if (fVerbosity > 0)
      std::cerr << "Table::CreateFromDB command failed: " 
	   << PQerrorMessage(fConnection) << std::endl;
    PQclear(res);
    fExistsInDB = false;  // set this to false just in case
    CloseConnection();
    return false;
  }

  nRow = PQntuples(res);

  for (int i=0; i<nRow; ++i) {
    std::string cname = std::string(PQgetvalue(res,i,0));
    std::string ctype = std::string(PQgetvalue(res,i,1));
    if (ctype == "smallint") ctype="short";
    else if (ctype == "double precision") ctype="double";
    else if (ctype == "boolean") ctype="bool";
    else if (ctype == "timestamp without time zone") ctype="timestamp";
    else if (ctype.substr(0,7) == "varchar") 
      ctype = "varchar" + ctype.substr(8,ctype.find(')')-1);
    
    // check if this column is "auto_incr"
    if (ctype == "integer") {
      cmd = "SELECT pg_get_serial_sequence(\'" + stName + 
	"\',\'" + cname + "\')";
      PGresult* res2 = PQexec(fConnection,cmd.c_str());
      int nRow2 = PQntuples(res2);
      for (int j=0; j<nRow2; ++j) {
	std::string tStr = std::string(PQgetvalue(res2,j,0));
	if (tStr != "") ctype = "auto_incr";
      }
      PQclear(res2);
    }

    // now create Column based on this info
    dBColumn_t c;
    c.setName(cname);
    c.setColType(ctype);
    
    if (find(pkeyList.begin(),pkeyList.end(),cname) != pkeyList.end()) {
      c.setNull(false);
      pkey.addDBCol(dBColumn_t(c));
    }
    fTable->addDBColumn(dBColumn_t(c));
    fCol.insert(fCol.begin(),Column(c));    

    XMLCol[cname] = &(fTable->getDBColumnReference(i));

    if (cname == "inserttime") addInsertTime = true;
    if (cname == "insertuser") addInsertUser = true;
    if (cname == "updatetime") addUpdateTime = true;
    if (cname == "updateuser") addUpdateUser = true;
  }

  PQclear(res);

  if (!hasConn) CloseConnection();

  // now set the dB command cache file name
  string dirName;
  char* tmpStr = getenv("NOVADBCACHEDIR");
  if (tmpStr) 
    dirName = tmpStr;
  else {
    tmpStr = getenv("PWD");
    if (tmpStr)
      dirName = tmpStr;
    else 
      dirName = "/";
  }

  fDBCacheFile = dirName + "/" + "." + Name() + ".cache";

  fTestedXML=true;
  fHasValidXML=true;
    
  return true;
}

//************************************************************

void Table::CreateFromXML(const char* tname)
{
  std::string fname;
  std::string errStr;
  fPassword="";

  if (! novadaq::LocationUtils::findFile(tname, fname)) {
    errStr = "Table::CreateFromXML: unable to find the file " + string(tname);
    throw std::runtime_error(errStr);
  }

  try {
    // 10-Aug-2011, KAB - added class-wide locking of this section
    // of code to prevent multiple threads from reading in a table
    // definition file at the same time. Without this, it's possible
    // for the code to crash since the Xerces-C library (which is
    // used by the XSD code) is not thread safe.
    boost::mutex::scoped_lock sl(_xsdLock);

    fTable = dBTable_(fname.c_str(),::xml_schema::flags::dont_validate);
  }
  catch (xsd::cxx::tree::parsing<char>& excpt) {
    errStr = "Table::CreateFromXML: " + string(excpt.what());
    throw std::runtime_error(errStr);
  }
  catch (xsd::cxx::tree::expected_element<char>& excpt) {
    errStr = "Table::CreateFromXML: " + string(excpt.what());
    throw std::runtime_error(errStr);
  }

  // convert table name to all lower case for convenience
  std::string tName = this->Name();
  boost::to_lower(tName);
  fTable->setTableName(tName);

  // set the DB host and name
  fDBHost = std::string(fTable->getDbHost());
  fDBName = std::string(fTable->getDbName());

  // set the table type.  By default, a table is of type "generic", 
  // eg. no default validity context
  if (fTable->hasTableType()) {
    std::string ttype = fTable->getTableType();
    boost::to_lower(ttype);
    int ittype = kNTableType;
    if (ttype == "generic") ittype = kGenericTable;
    if (ttype == "hardware") ittype = kHardwareTable;
    if (ttype == "validity") ittype = kValidityTable;
    if (! this->SetTableType(ittype)) {
      errStr = "Table::CreateFromXML: " + fTable->getTableName() + " unrecognized table type " + ttype;
      throw std::runtime_error(errStr);
    }
  }

  if (fTable->hasSchema()) {
    fSchema = std::string(fTable->getSchema());
    boost::to_lower(fSchema);
  }

  if (! fTable->hasDBPKey() && fTableType != kValidityTable ) {
    errStr = "Table::CreateFromXML: " + fTable->getTableName() + " has no primary key defined";
    throw std::runtime_error(errStr);
  }

  // now add columns where necessary
  dBColumn_t c;

  dBTable::dBPKey_type pkey;
  if (fTable->hasDBPKey())
    pkey = dBTable::dBPKey_type(fTable->getDBPKey());

  fTable->setDBPKey(pkey);

  addInsertTime = false;
  if (fTable->hasAddInsertTime()) 
    addInsertTime = fTable->getAddInsertTime();

  if (addInsertTime)  {
    c.setNull(false);
    c.setName("inserttime");
    c.setColType("timestamp");
    fTable->addDBColumn(dBColumn_t(c));
    addInsertTime = true;
  }

  addInsertUser = false;
  if (fTable->hasAddInsertUser()) 
    addInsertUser = fTable->getAddInsertUser();
  if (addInsertUser) {
    c.setNull(false);
    c.setName("insertuser");
    c.setColType("string");
    fTable->addDBColumn(dBColumn_t(c));
    addInsertUser = true;
  }

  addUpdateTime = false;
  if (fTable->hasAddUpdateTime()) 
    addUpdateTime = fTable->getAddUpdateTime();
  if (addUpdateTime) {
    c.setNull(true);
    c.setName("updatetime");
    c.setColType("timestamp");
    fTable->addDBColumn(dBColumn_t(c));
    addUpdateTime = true;
  }

  addUpdateUser = false;
  if (fTable->hasAddUpdateUser()) 
    addUpdateUser = fTable->getAddUpdateUser();
  if (addUpdateUser) {
    c.setNull(true);
    c.setName("updateuser");
    c.setColType("string");
    fTable->addDBColumn(dBColumn_t(c));
    addUpdateUser = true;
  }

  // add necessary columns for Validity type tables (channel and timestamp)
  if (TableType() == kValidityTable) {
    c.setName("channel");
    c.setColType("int");
    fTable->addDBColumn(dBColumn_t(c));
    c.setName("tv");
    //    c.setColType("timestamp");
    c.setColType("long");
    fTable->addDBColumn(dBColumn_t(c));
  }

  fXMLOverDB = false;
  if (fTable->hasXmlOverDB())
    fXMLOverDB = fTable->getXmlOverDB();

  fCol.clear();

  for (int i=0; i<fTable->getDBColumnCount(); ++i) {
    Column col(fTable->getDBColumn(i));
    string cName = col.Name(); // convert name to all lower case
    boost::to_lower(cName);
    col.SetName(cName);
    fCol.push_back(col);
    XMLCol[cName] = &(fTable->getDBColumnReference(i));
  }

  fPKeyList.clear();
  if (fTable->hasDBPKey()) {
    int npkey = fTable->getDBPKey().getDBColCount();
    for (int j=0; j<npkey; ++j) {
      string pkey = fTable->getDBPKey().getDBCol(j).getName();
      if (this->GetCol(pkey))
	fPKeyList.push_back(this->GetCol(pkey));
    }
  }

  fRow.clear();
  for (int i=0; i<fTable->getDBRowCount(); ++i) {
    Row r(fCol, fTable->getDBRow(i));
    fRow.push_back(r);
  }

  fIgnoreDB = false;
  if (fTable->hasIgnoreDB()) 
    fIgnoreDB = fTable->getIgnoreDB();

  //  GetConnectionInfo();
  
  char* tmpStr = getenv("NOVADBNCNXTRIES");
  if (tmpStr) {
    int nTry = atoi(tmpStr);
    if (nTry > 0)
      fMaxConnAttempts = nTry;
  }

  // now set the dB command cache file name
  string dirName;
  tmpStr = getenv("NOVADBCACHEDIR");
  if (tmpStr) 
    dirName = tmpStr;
  else {
    tmpStr = getenv("PWD");
    if (tmpStr)
      dirName = tmpStr;
    else 
      dirName = "/";
  }

  fDBCacheFile = dirName + "/" + "." + Name() + ".cache";
    
}

//************************************************************

bool Table::SetTableType(int t)
{
  if (t < 0 || t >= kNTableType) return false;
  fTableType = t;
  return true;
}

//************************************************************
void Table::Reset()
{
  fConnection = 0;
  fHasConnection = 0;
  fXMLOverDB = false;
  /*
  addInsertTime = false;
  addInsertUser = false;
  addUpdateTime = false;
  addUpdateUser = false;
  */
  fPKeyList.clear();
  fDistinctCol.clear();
  fVerbosity = 0;
  fDescOrder = true; 
  fSelectLimit = 0;
  fSelectOffset = 0;
  ClearValidity();
  
  char* tmpStr = getenv("NOVADBVERB");
  if (tmpStr) {
    fVerbosity = atoi(tmpStr);
  }
  
}

//************************************************************
void Table::ClearValidity() 
{ 
  fValidityStart.clear();
  fValidityEnd.clear(); 
  fValiditySQL = "";
  fValidityChanged=true;
}

//************************************************************
bool Table::SetDetector(std::string det)
{
  // check that detector string is valid
  if (novadaq::cnv::DetInfo::GetId(det) == novadaq::cnv::kUNKNOWN_DET) {
    std::cerr << "Error in Table::SetDetector: Unknown detector name \'" 
	      << det << "\'" << std::endl;
    return false;
  }
  
  fDetector = det;

  if (fTableType != kHardwareTable)
    fSchema = det;
  else
    fSchema = "public";

  boost::to_lower(fSchema);
    
  return true;
}

//************************************************************
bool Table::SetDetector(int idet)
{
  // check that detector id is valid
  if (idet <= novadaq::cnv::kUNKNOWN_DET || idet >= novadaq::cnv::kNDetector) {
    std::cerr << "Error in Table::SetDetector: Unknown detector id" << idet
	      << std::endl;
    return false;
  }
  
  fDetector = novadaq::cnv::DetInfo::GetName(idet);

  if (fTableType != kHardwareTable)
    fSchema = fDetector;
  else
    fSchema = "public";

  boost::to_lower(fSchema);
  
  return true;
}

//************************************************************
bool Table::GetDetector(std::string& det) const
{
  if (fDetector == "") return false;
  det = fDetector;

  return true;
}

//************************************************************
bool Table::GetDetector(int& idet) const
{
  if (idet <= novadaq::cnv::kUNKNOWN_DET || idet >= novadaq::cnv::kNDetector) {
    std::cerr << "Error in Table::GetDetector: Unknown detector id" << idet
	      << std::endl;
    return false;
  }
  idet = novadaq::cnv::DetInfo::GetId(fDetector);

  return true;
}

//************************************************************
void Table::SetDBInfo(std::string name, std::string host, std::string port,
		      std::string user)
{
  SetDBName(name);
  SetDBHost(host);
  SetDBPort(port);
  SetUser(user);
}

//************************************************************
void Table::SetDBInfo(const char* name, const char* host, const char* port,
		      const char* user)
{
  SetDBName(name);
  SetDBHost(host);
  SetDBPort(port);
  SetUser(user);
}

//************************************************************
const Column* Table::GetCol(string& cname) 
{
  unsigned int i=0;
  for ( ; i < fCol.size(); ++i) 
    if (fCol[i].Name() == cname) break;
  
  if (i >= fCol.size()) return 0;

  return &fCol[i];
}

//************************************************************
vector<string> Table::GetColNames()
{
  vector<string> nameList;

  for (unsigned int i=0; i<fCol.size(); ++i)
    nameList.push_back(fCol[i].Name());

  return nameList;
}

//************************************************************
void Table::SetTolerance(std::string& cname, float t)
{
  unsigned int i=0;
  for ( ; i < fCol.size(); ++i) 
    if (fCol[i].Name() == cname) break;
  
  if (i >= fCol.size()) return;

  fCol[i].SetTolerance(t);
  
}
//************************************************************
void Table::SetToleranceType(std::string& cname, int t)
{
  unsigned int i=0;
  for ( ; i < fCol.size(); ++i) 
    if (fCol[i].Name() == cname) break;
  
  if (i >= fCol.size()) return;

  fCol[i].SetToleranceType(t);

}

//************************************************************
float Table::Tolerance(std::string& cname)
{
  unsigned int i=0;
  for ( ; i < fCol.size(); ++i) 
    if (fCol[i].Name() == cname) break;
  
  if (i >= fCol.size()) return 0.;

  return fCol[i].Tolerance();

}

//************************************************************
float Table::ToleranceType(std::string& cname)
{
  unsigned int i=0;
  for ( ; i < fCol.size(); ++i) 
    if (fCol[i].Name() == cname) break;
  
  if (i >= fCol.size()) return kUnknown;

  return fCol[i].ToleranceType();

}

//************************************************************
void Table::PrintColumns()
{
  std::cout << std::endl;
  int i = 0;
  vector<int> len(0);
  int tlen;
  int sumlen = 0;

  for ( ; i<this->NCol(); ++i) {
    tlen = this->GetCol(i)->Name().length();
    if ((int)this->GetCol(i)->Type().length() > tlen)
      tlen = this->GetCol(i)->Type().length();
    len.push_back(tlen);
    sumlen += tlen;
  }

  int nsp = 0;
  i = 0;
  int j = 0;
  while (i < this->NCol()) {
    for ( ; i<this->NCol() && nsp<78; ++i)
      nsp += len[i] + 1;

    for (int k=0; k<nsp; ++k) {
      std::cout << "_" ;
    }
    std::cout << std::endl;

    int j_save = j;
    for (; j<i; ++j)
      std::cout << "|" << setw(len[j]) << left << this->GetCol(j)->Name();
    std::cout << "|" << std::endl;
    
    for (int k=0; k<nsp; ++k) {
      std::cout << "-" ;
    }
    std::cout << std::endl;

    j = j_save;
    for ( ; j<i; ++j)
      std::cout << "|" << setw(len[j]) << left << this->GetCol(j)->Type();
    std::cout << "|" << std::endl;

    for (int k=0; k<nsp; ++k) {
      std::cout << "-" ;
    }
    std::cout << std::endl;

    nsp = 0;
  }

}

//************************************************************
void Table::AddRow(const Row* row) 
{
  if (!row) return;

  Row r2(*row);

  for (unsigned int i=0; i<fCol.size(); ++i) {
    if (r2.Col(i).Name() == "inserttime" ||
	r2.Col(i).Name() == "insertuser" ||
	r2.Col(i).Name() == "updatetime" ||
	r2.Col(i).Name() == "updateuser" ) continue;
    /*
    if (TableType() == "runsubrun") {
      r2.Col("minrun")->Set(fRunVld->Value());
      r2.Col("minsubrun")->Set(fSubrunVld->Value());
      r2.Col("maxrun")->Set(fRunVld->Value());
      r2.Col("maxsubrun")->Set(fSubrunVld->Value());
    }
    if (TableType() == "timestamp") {
      r2.Col("mintime")->Set(fTSVld->Value());
      r2.Col("maxtime")->Set(fTSVld->Value());
    }
    */
    if (!r2.Col(i).CanBeNull())
      if (r2.Col(i).IsNull()) 
	fNullList.push_back(pair<int,int>(fRow.size(),i));
  }

  fRow.push_back(r2); 

}

//************************************************************
void Table::AddRow(const Row& row) 
{
  AddRow(&row);
}

//************************************************************
bool Table::RemoveRow(int i) 
{
  if (i < 0) return false;

  unsigned int j = i;
  
  if (j >= fRow.size()) return false;

  unsigned int kEnd = fNullList.size();
  for (unsigned int k=0; k<kEnd; ++k) 
    if (fNullList[k].first == i) {
      fNullList.erase(fNullList.begin()+k);
      kEnd = fNullList.size();
    }
  
  fRow.erase(fRow.begin()+j);

  return true;
}

//************************************************************
bool Table::CheckForNulls()
{
  //  if (TableType() == "runsubrun" || TableType() == "timestamp") return true;

  bool isOk = fNullList.empty();

  if (!isOk) // print out list of null columns
    for (unsigned int i=0; i<fNullList.size(); ++i) 
      if (fVerbosity>0)
	std::cerr << fCol[fNullList[i].second].Name() << " is NULL in row " 
		  << fNullList[i].first << std::endl;

  return isOk;

}

//************************************************************
Row* const Table::GetRow(int i)
{
  if (i >= 0 && i < (int)fRow.size())
    return &fRow[i];
  else
    return 0;
}

//************************************************************
Row* const Table::GetRowByPKey(vector<string>& pkeyval)
{
  if (pkeyval.size() != fPKeyList.size()) {
    std::cerr << "Table::GetRowByPKey: too few pkeys provided!" << std::endl;
    return 0;
  }
  
  vector<string> pkeyname;
  for (unsigned int i=0; i<fPKeyList.size(); ++i) 
    pkeyname.push_back(GetPKeyCol(i)->Name());

  for (unsigned int i=0; i<fRow.size(); ++i) {
    unsigned int j=0;
    for (; j<pkeyval.size(); ++j) {
      if (fRow[i].Col(pkeyname[j])->Value() != pkeyval[j]) break;
    }
    
    if (j == pkeyval.size()) {
      return &fRow[i];
    }
  }  
  return 0;
}

//************************************************************
vector<Row*> const Table::GetRowByColSet(vector<Column>& cols)
{
  vector<Row*> rows(0);

  vector<unsigned int> colidx(cols.size());
  vector<string> colval(cols.size());

  // create index to cols
  for (unsigned int i=0; i<cols.size(); ++i) {
    string colStr = cols[i].Name();
    for (unsigned int j=0; j<fCol.size(); ++j) {
      if (fCol[j].Name() == colStr) {
	colidx[i] = j;
	cols[i].Get(colval[i]);
	break;
      }
    }
  }

  string vStr;
  for (unsigned int i=0; i<fRow.size(); ++i) {
    bool isMatch = true;
    for (unsigned int j=0; j<cols.size(); ++j) {
      unsigned int k=colidx[j];      
      if (fRow[i].Col(k).Value() !=  colval[j]) {
	isMatch = false;
	break;
      }
    }
    if (isMatch)  rows.push_back(&fRow[i]);
  }

  return rows;
  
}

//************************************************************

void Table::CacheDBCommand(string cmd) 
{
  ofstream fout;
  fout.open(fDBCacheFile.c_str(),ios_base::app);

  fout << cmd << std::endl;

  fout.close();  
}

//************************************************************
// This function returns true if all column names in the XML
// table definition match those in the dB
bool Table::CheckValidXMLDef()
{
  if (fIgnoreDB) return false;

  if (fTestedXML) return fHasValidXML;

  if (fTableType != kHardwareTable)
    if (fDetector == "") {
      std::cerr << "Table::CheckValidXMLDef: Detector Validity not set!" << std::endl;
      return false;
    }
  
  // make a connection to the dB if there isn't one already
  bool hasConn = fHasConnection;
  if (! fHasConnection) {
    if (! GetConnection()) {
      std::cerr << "Table::CheckValidXMLDef: No connection to the database!" 
		<< std::endl;
      return false;
    }
    hasConn = false;
  }

  if (!ExistsInDB()) {
    std::cerr << "Table::CheckValidXMLDef: Table not found in database!" 
	      << std::endl;
    CloseConnection();
    return false;
  }
  
  string cmd;
  PGresult* res;

  cmd = "SELECT column_name from information_schema.columns where table_schema=\'" 
    + Schema() + "\' and table_name=\'" + Name() + "\'";

  res = PQexec(fConnection, cmd.c_str());
  if (PQresultStatus(res) != PGRES_TUPLES_OK) {
    std::cerr << "Table::CheckValidXML: Select failed: %" 
	      << PQerrorMessage(fConnection) << std::endl;
    PQclear(res);
    CloseConnection();
    return false;
  }
  
  // now cache rows
  //  int nCol = PQnfields(res);
  int nRow = PQntuples(res);

  bool isOk = true;

  for (unsigned int i=0; i<fCol.size(); ++i) {
    string cName = fCol[i].Name();

    int j=0;
    for (; j<nRow; ++j) {
      string cName2 = PQgetvalue(res,j,0);
      if (cName2 == cName) break;
    }
    
    if (j == nRow) {
      isOk = false;
      break;
    }
  }

  PQclear(res);
  
  // close connection to the dB if necessary
  if (! hasConn) CloseConnection();

  fTestedXML = true;
  fHasValidXML = isOk;

  return fHasValidXML;
  
}

//************************************************************
bool Table::LoadFromDB()
{
  if (fIgnoreDB) return false;

  if (fSchema == "undef") {
    std::cerr << "Table::LoadFromDB: Detector not set!  Table::SetDetector()"
	      << " must be called first!" << std::endl;
    return false;
  }

  if (!fValidityChanged) return true;

  // make a connection to the dB if there isn't one already
  bool hasConn = fHasConnection;
  if (! fHasConnection) {
    GetConnection();
    hasConn = false;
  }

  if (!fConnection) {
    std::cerr << "Table::LoadFromDB: No connection to the database!" << std::endl;
    return false;
  }

  if (!ExistsInDB()) {
    std::cerr << "Table::LoadFromDB: Table \"" << Name() 
	 << "\" not found in database!" << std::endl;
    CloseConnection();
    return false;
  }
  
  if (! CheckValidXMLDef()) {
    std::cerr << "Table::LoadFromDB: Found mismatch between XML table definition ";
    std::cerr << "file and table defintion from the database for table " 
	 << Name() << "!  Aborting read..." << std::endl;
    return false;
  }

  string cmd;
  cmd.clear();
  PGresult* res;

  /*
  cmd = "SET search_path TO " + fSchema;
  res = PQexec(fConnection, cmd.c_str());
  if (PQresultStatus(res) != PGRES_COMMAND_OK) {
    std::cerr << "\'" << cmd << "\' command failed" << std::endl;
    PQclear(res);
    CloseConnection();
    return false;
  }	
  PQclear(res);
  cmd.clear();  
  */

  std::ostringstream outs;
  outs << "BEGIN";
  res = PQexec(fConnection, outs.str().c_str());
  if (PQresultStatus(res) != PGRES_COMMAND_OK) {
    std::cerr << "BEGIN command failed: " << PQerrorMessage(fConnection) << std::endl;
    PQclear(res);
    CloseConnection();
    return false;
  }
  
  PQclear(res);

  outs.str("");
  outs << "DECLARE myportal CURSOR FOR SELECT ";
  if (!fDistinctCol.empty()) {
    outs << "DISTINCT ON (";    
    if (! fDistinctCol.empty()) {
      for (unsigned int i=0; i<fDistinctCol.size(); ++i) {
	outs << fDistinctCol[i]->Name();
	if (i<(fDistinctCol.size()-1)) outs << ", ";
      }    
    }    
    outs << ") ";
  }
  
  outs << "* from ";
  outs << Schema() << "." << Name();

  if (! fValidityStart.empty() || fValiditySQL != "" ) {
    outs << " WHERE " << fValiditySQL;
    if (fValiditySQL != "" && !fValidityStart.empty()) outs << " and ";
    
    for (unsigned int i=0; i<fValidityStart.size(); ++i) {
      bool isEqualTo = (fValidityStart[i].Value() == fValidityEnd[i].Value());
      
      outs << fValidityStart[i].Name();
      if (!isEqualTo)
	outs << ">=";
      else
	outs << "=";
      
      outs << fValidityStart[i];
      
      if (!isEqualTo) {
	outs << " and ";
	outs << fValidityEnd[i].Name() + "<=";
	outs << fValidityEnd[i];
      }
      
      if (i < (fValidityStart.size()-1)) outs << " and ";
    }
  }

  if (!fDistinctCol.empty() || !fOrderCol.empty()) {
    outs << " ORDER BY ";

    if (!fDistinctCol.empty()) {
      for (unsigned int i=0; i<fDistinctCol.size(); ++i) {      
	outs << fDistinctCol[i]->Name();
	if (i<(fDistinctCol.size()-1)) outs << ", ";      
      }    
    }
    
    if (!fOrderCol.empty()) {
      for (unsigned int i=0; i<fOrderCol.size(); ++i) {
	outs << fOrderCol[i]->Name();
	if (i<(fOrderCol.size()-1)) outs << ", ";
      }
    }
    
    if (fDescOrder)
      outs << " DESC";
    else
      outs << " ASC";
  }
  
  if (fSelectLimit>0) {
    outs << " LIMIT " << boost::lexical_cast<string>(fSelectLimit);
  }

  if (fSelectOffset>0) {
    outs << " OFFSET " << boost::lexical_cast<string>(fSelectOffset);
  }

  if (fVerbosity > 0)
    std::cerr << "Table::LoadFromDB: Executing PGSQL command: \n\t" << outs.str() << std::endl;
 
  res = PQexec(fConnection,outs.str().c_str());

  if (PQresultStatus(res) != PGRES_COMMAND_OK) {
    std::cerr << "DECLARE CURSOR failed: " << PQerrorMessage(fConnection) << std::endl;
    PQclear(res);
    CloseConnection();
    return false;
  }
  PQclear(res);
  
  res = PQexec(fConnection, "FETCH ALL in myportal");
  if (PQresultStatus(res) != PGRES_TUPLES_OK) {
    std::cerr << "FETCH ALL failed: %" << PQerrorMessage(fConnection) << std::endl;
    PQclear(res);
    CloseConnection();
    return false;
  }
  
  // now cache rows
  int nRow = PQntuples(res);
  vector<int> colMap(fCol.size());

  for (unsigned int i=0; i<fCol.size(); ++i) {
    colMap[i] = PQfnumber(res,fCol[i].Name().c_str());
  }

  int k;
  Row* row = NewRow();
  row->SetInDB();
  std::vector<std::string> hashVec;
  for (int i=0; i < nRow; i++) {    
    row->Clear();
    for (unsigned int j=0; j < fCol.size(); j++) {
      k = colMap[j];
      if (k >= 0) {
	if (! PQgetisnull(res,i,k)) {
	  string vstr = PQgetvalue(res,i,k);
	  bool isAutoIncr = (row->Col(j).Type() == "autoincr");	  
	  row->Col(j).Set(vstr,isAutoIncr);
	}
	else
	  row->Col(j).SetNull(true);
      }
    }
    AddRow(row);
  }
  delete row;
  
  PQclear(res);
  
  /* close the portal ... we don't bother to check for errors ... */
  res = PQexec(fConnection, "CLOSE myportal");
  PQclear(res);
  
  /* end the transaction */
  res = PQexec(fConnection, "END");
  PQclear(res);

  // close connection to the dB if necessary
  if (! hasConn) CloseConnection();

  fValidityChanged = false;

  return true;
}

//************************************************************
bool Table::LoadFromCSV(std::string fname, bool update)
{
  string tname;
  if (! novadaq::LocationUtils::findFile(fname, tname)) {
    std::cerr << "Cannot find " << fname << std::endl;
    return false;
  }
  
  std::cout << "Reading " << tname << std::endl;
  
  ifstream fin;
  fin.open(tname.c_str());
  if (!fin.is_open() || !fin.good()) {
    std::cerr << "Could not open " << tname << std::endl;
    return false;
  }

  unsigned int nReqCol = 0;
  for (unsigned int j=0; j<fCol.size(); ++j)
    if (fCol[j].Name() == "inserttime" ||
	fCol[j].Name() == "insertuser" ||
	fCol[j].Name() == "updatetime" ||
	fCol[j].Name() == "updateuser" ) continue;
    else
      nReqCol++;
  
  using namespace boost;

  string s;
  while (!fin.eof()) {
    getline(fin,s);
    if (fin.eof()) break;

    tokenizer<escaped_list_separator<char> >  tok(s);
    tokenizer<escaped_list_separator<char> >::iterator itr=tok.begin();
    tokenizer<escaped_list_separator<char> >::iterator itrEnd=tok.end();
    
    unsigned int sz = 0;
    for (; itr != itrEnd; ++itr) ++sz;

    itr = tok.begin();

    // require the correct number of commas in the line if we are to 
    // insert it into the table

    if (sz >= nReqCol && sz <= fCol.size()) {
      Row* r = NewRow();
      for (int j=0; itr != itrEnd; ++itr,++j) {
	string value = *itr;
	boost::trim(value);
	if (value.find("x") != std::string::npos) {
            if (r->Col(j).Type() == "bigint" ||
                r->Col(j).Type() == "long") {
                try {
                    std::istringstream iss(value);
                    uint64_t ulongValue;
                    iss >> std::hex >> ulongValue;
                    int64_t longValue = (int64_t) ulongValue;
                    value = boost::lexical_cast<string>(longValue);
                }
                catch (...) {
                    // simply let "value" remain unchanged
                }
            }
            else if (r->Col(j).Type() == "int") {
                try {
                    std::istringstream iss(value);
                    uint32_t uintValue;
                    iss >> std::hex >> uintValue;
                    int32_t intValue = (int32_t) uintValue;
                    value = boost::lexical_cast<string>(intValue);
                }
                catch (...) {
                    // simply let "value" remain unchanged
                }
            }
            else if (r->Col(j).Type() == "short") {
                try {
                    std::istringstream iss(value);
                    uint16_t ushortValue;
                    iss >> std::hex >> ushortValue;
                    int16_t shortValue = (int16_t) ushortValue;
                    value = boost::lexical_cast<string>(shortValue);
                }
                catch (...) {
                    // simply let "value" remain unchanged
                }
            }
	}

	bool isAutoIncr = (r->Col(j).Type() == "autoincr");
	if (value.empty()) {
	  r->Col(j).SetNull(true);
	  value = "NULL";
	}
	if (value == "\"")
	  value = "";

	if (!update)
	  r->Col(j).Set(value,isAutoIncr);
	else{
	  string name = r->Col(j).Name();
	  r->Update(name,value);
	}
      }

      bool isValid = true;
      if (! fValidityStart.empty()) {
	for (unsigned int i=0; i<fValidityStart.size(); ++i) {
	  const Column* c = r->Col(fValidityStart[i].Name());
	  if (fValidityStart[i].Type() == "bool") 
	    isValid = (c->Value() == fValidityStart[i].Value());
	  else 
	    isValid = ((*c >= fValidityStart[i]) && (*c <= fValidityEnd[i]));
	  if (!isValid) break;
	}
      }

      if (isValid) AddRow(r);
      delete r;
    }
    else {
      std::cerr << "Table::LoadFromCSV: ignoring line, wrong number of entries, "
           << "expected at least " << nReqCol << ", found " << sz << "."
	   << std::endl;
    }
  }
  
  fin.close();
  
  return true;
}

//************************************************************
bool Table::UpdateFromCSV(std::string fname)
{
  return LoadFromCSV(fname,true);
  
}

//************************************************************
bool Table::LoadFromSSV(std::string fname, bool update)
{
  string tname;
  if (! novadaq::LocationUtils::findFile(fname, tname)) {
    std::cerr << "Cannot find " << fname << std::endl;
    return false;
  }
  
  std::cout << "Reading " << tname << std::endl;
  
  ifstream fin;
  fin.open(tname.c_str());
  if (!fin.is_open() || !fin.good()) {
    std::cerr << "Could not open " << tname << std::endl;
    return false;
  }

  using namespace boost;

  char_separator<char>ws(" ");

  string s;
  while (!fin.eof()) {
    getline(fin,s);
    if (fin.eof()) break;

    tokenizer<char_separator<char> >  tok(s,ws);
    tokenizer<char_separator<char> >::iterator itr=tok.begin();
    tokenizer<char_separator<char> >::iterator itrEnd=tok.end();

    unsigned int sz = 0;
    for (; itr != itrEnd; ++itr) {
      ++sz;
    }

    itr = tok.begin();

    // require the correct number of commas in the line if we are to 
    // insert it into the table
    if (sz == fCol.size()) {
      Row* r = NewRow();
      for (int j=0; itr != itrEnd; ++itr,++j) {
	string value = *itr;
	boost::trim(value);
	if (value.find("x") != std::string::npos) {
            if (r->Col(j).Type() == "bigint" ||
                r->Col(j).Type() == "long") {
                try {
                    std::istringstream iss(value);
                    uint64_t ulongValue;
                    iss >> std::hex >> ulongValue;
                    int64_t longValue = (int64_t) ulongValue;
                    value = boost::lexical_cast<string>(longValue);
                }
                catch (...) {
                    // simply let "value" remain unchanged
                }
            }
            else if (r->Col(j).Type() == "int") {
                try {
                    std::istringstream iss(value);
                    uint32_t uintValue;
                    iss >> std::hex >> uintValue;
                    int32_t intValue = (int32_t) uintValue;
                    value = boost::lexical_cast<string>(intValue);
                }
                catch (...) {
                    // simply let "value" remain unchanged
                }
            }
            else if (r->Col(j).Type() == "short") {
                try {
                    std::istringstream iss(value);
                    uint16_t ushortValue;
                    iss >> std::hex >> ushortValue;
                    int16_t shortValue = (int16_t) ushortValue;
                    value = boost::lexical_cast<string>(shortValue);
                }
                catch (...) {
                    // simply let "value" remain unchanged
                }
            }
	}

	bool isAutoIncr = (r->Col(j).Type() == "autoincr");
	if (!update)
	  r->Col(j).Set(value,isAutoIncr);
	else{
	  string name = r->Col(j).Name();
	  r->Update(name,value);
	}
      }

      bool isValid = true;
      if (! fValidityStart.empty()) {
	for (unsigned int i=0; i<fValidityStart.size(); ++i) {
	  const Column* c = r->Col(fValidityStart[i].Name());
	  if (fValidityStart[i].Type() == "bool") 
	    isValid = (c->Value() == fValidityStart[i].Value());
	  else 
	    isValid = ((*c >= fValidityStart[i]) && (*c <= fValidityEnd[i]));
	  if (!isValid) break;
	}
      }

      if (isValid) AddRow(r);
      delete r;
    }
    else {
      std::cerr << "Table::LoadFromSSV: ignoring line, wrong number of entries, "
           << "expected " << fCol.size() << ", found " << sz << "."
	   << std::endl;
    }
  }
  
  fin.close();
  
  return true;
}

//************************************************************
bool Table::UpdateFromSSV(std::string fname)
{
  return LoadFromSSV(fname,true);
  
}

//************************************************************
bool Table::GetConnectionInfo()
{
  char* tmpStr;

  if (!fIgnoreEnvVar) {
    // now check environment variables, which will override any previous settings
    if (fTableType == kHardwareTable) {
      tmpStr = getenv("NOVAHWDBHOST");
      if (tmpStr) 
	fDBHost = tmpStr;
      tmpStr = getenv("NOVAHWDBNAME");
      if (tmpStr)
	fDBName = tmpStr;
      tmpStr = getenv("NOVAHWDBPORT");
      if (tmpStr) 
	fDBPort = tmpStr;
      tmpStr = getenv("NOVAHWDBUSER");
      if (tmpStr) 
	fUser = tmpStr;
    }
    else {
      tmpStr = getenv("NOVADBHOST");
      if (tmpStr) 
	fDBHost = tmpStr;
      tmpStr = getenv("NOVADBNAME");
      if (tmpStr)
	fDBName = tmpStr;
      tmpStr = getenv("NOVADBPORT");
      if (tmpStr) 
	fDBPort = tmpStr;
      tmpStr = getenv("NOVADBUSER");
      if (tmpStr) 
	fUser = tmpStr;
    }
  }

  if (fTableType == kDAQTable || fTableType == kDCSTable) {
    std::string bVar = "NOVAXXXX";      
    if (novadaq::cnv::DetInfo::GetId(fDetector) == 
	novadaq::cnv::kFARDET) {
      if (fTableType == kDAQTable)
	bVar = "NOVAFARDAQDB";
      else
	bVar = "NOVAFARDCSDB";
    }
    else if (novadaq::cnv::DetInfo::GetId(fDetector) == 
	     novadaq::cnv::kNEARDET) {
      if (fTableType == kDAQTable)
	bVar = "NOVANEARDAQDB";
      else
	bVar = "NOVANEARDCSDB";
    }
    std::string envVar = bVar + std::string("HOST");
    tmpStr = getenv(envVar.c_str());
    if (tmpStr) 
      fDBHost = tmpStr;
    
    envVar = bVar + std::string("NAME");
    tmpStr = getenv(envVar.c_str());
    if (tmpStr)
      fDBName = tmpStr;
    envVar = bVar + std::string("PORT");
    tmpStr = getenv(envVar.c_str());
    if (tmpStr) 
      fDBPort = tmpStr;
    
    envVar = bVar + std::string("USER");
    tmpStr = getenv(envVar.c_str());
    if (tmpStr) 
      fUser = tmpStr;
    
  }
  
  std::cout << "Connecting with: " << fDBHost << ":" << fDBPort 
	    <<  ":" << fDBName << ":" << fUser << std::endl;
  
  if (fUser == "") {
    tmpStr = getenv("USER");
    if (tmpStr) {
      fUser = tmpStr;
      std::cerr << "Table::GetConnectionInfo: DB User undefined.  Setting to \""
		<< fUser << "\"" << std::endl;
    }
    else {
      fUser = "nova_reader";
      std::cerr << "Table::GetConnectionInfo: DB User undefined.  " 
		<< "Setting to \"nova_reader\"" << std::endl;
    }  
  }
  if (fDBHost == "") {
    throw std::runtime_error("Table::GetConnectionInfo: DB HOST undefined.");
  }
  if (fDBName == "") {
    throw std::runtime_error("Table::GetConnectionInfo: DB NAME undefined.");
  }

  if (fDBPort == "")
    if (fTable->hasDbPort()) 
      fDBPort = fTable->getDbPort();

  return true;
}

//************************************************************
bool Table::GetConnection()
{
  if (fIgnoreDB) return false;

  try {
    GetConnectionInfo();
  }
  catch (std::runtime_error& e) {
    std::cerr << e.what() << std::endl;
    return false;
  }

  // now get the password file name for read-only access 
  // (nova_reader acct only)
  if (fUser == "nova_reader" || fUser == "nova_grid") 
    if (fPassword == "") {
      char* tmpStr = getenv("NOVADBPWDFILE");
      if (tmpStr) 
	if (! this->SetPasswordFile(tmpStr)) {
	  return false;
	}
    }
  
  if (!fConnection) {
    string cmd = "dbname = " + fDBName + " host = " + fDBHost + " user = " + fUser;
    if (fDBPort != "")
      cmd += " port = " + fDBPort;
    
    if (fPassword != "")
      cmd += " password = " + fPassword;

    fConnection = PQconnectdb(cmd.c_str());

    int nTry=0;
    int sleepTime = 2;
    while (PQstatus(fConnection) != CONNECTION_OK && 
	   ++nTry<fMaxConnAttempts) {
      std::cerr << "Connection to " << fDBHost << ":" 
		<< fDBName << " failed: " 
		<< PQerrorMessage(fConnection) << std::endl;
      
      CloseConnection();
      fConnection = PQconnectdb(cmd.c_str());
      sleep(sleepTime);
      sleepTime = int(60./(1+29*exp(-0.5*nTry)));
    }
    if (nTry == fMaxConnAttempts) {
      std::cerr << "Too many attempts to connect to " << fDBHost
		<< ", giving up." << std::endl;
      CloseConnection();
      return false;
    }
    fHasConnection = true;
    if (fVerbosity > 0)
      std::cout << "Got new connection" << std::endl;
  }

  return true;
}

//************************************************************
bool Table::CloseConnection()
{
  if (fConnection) {
    PQfinish(fConnection);
    if (fVerbosity > 0)
      std::cout << "Closed connection" << std::endl;
  }

  fConnection = 0;
  fHasConnection = false;

  return true;
}

//************************************************************
bool Table::SetRole(string role)
{  
  fRole = role;
  return true;
}

//************************************************************
bool Table::SetPasswordFile(const char* fname)
{
  string fNameStr = "";

  if (fname == 0) {
    char* tmpStr = getenv("NOVADBPWDFILE");
    if (tmpStr)
      fNameStr = tmpStr;
    else {
      std::cerr << "NOVADBPWDFILE env. variable is not set, disabling "
	   << "password-access to the dB." << std::endl;
      fPassword = "";
      return false;
    }
  }
  else
    fNameStr = fname;

  ifstream fin;
  fin.open(fNameStr.c_str());
  if (!fin.is_open() || !fin.good()) {
    std::cerr << "Could not open password file " << fNameStr 
	 << ".  Disabling password-access to the dB." << std::endl;
    return false;
  }
  else {
    fin >> fPassword;
    fin.close();
  }

  return true;
}

//************************************************************
bool Table::ExistsInDB()
{
  if (fIgnoreDB) return false;

  if (fTestedExists) return fExistsInDB;

  std::string tname = this->Name();

  fTestedExists = true;

  bool hasConn = fHasConnection;
  if (! fHasConnection) {
    GetConnection();
    hasConn = false;
  }    

  std::string cmd = "SELECT tablename FROM pg_tables WHERE schemaname=\'" +
    fSchema + "\'";
  //  std::cout << tname << ": " << cmd << std::endl;
  PGresult* res = PQexec(fConnection,cmd.c_str());
  
  if (PQresultStatus(res) != PGRES_TUPLES_OK) {
    if (fVerbosity > 0)
      std::cerr << "Table::ExistsInDB command failed: " 
	   << PQerrorMessage(fConnection) << std::endl;
    PQclear(res);
    fExistsInDB = false;  // set this to false just in case
    CloseConnection();
    return false;
  }

  bool retVal = false;
  int nRow = PQntuples(res);
  
  for (int i=0; i<nRow; ++i) {
    //    std::cout << string(PQgetvalue(res,i,0)) << std::endl;

    if (string(PQgetvalue(res, i, 0)) == tname) {
      retVal = true;
      break;
    }
  }

  PQclear(res);

  if (!hasConn) CloseConnection();

  fExistsInDB = true;
  return retVal;
  
}

//************************************************************
bool Table::GetCurrSeqVal(std::string col, int64_t& iseq)
{
  if (fIgnoreDB) return false;

  bool hasConn = fHasConnection;
  if (! fHasConnection) {
    GetConnection();
    hasConn = false;
  }    

  // now get current sequence value:

  string cmd = "SELECT last_value FROM "; 
  cmd += Schema() + "." + Name();
  cmd += "_" + col + "_seq";

  if (fVerbosity > 0)
    std::cerr << "Table::GetCurrSeqVal: Executing PGSQL command: \n\t" 
	 << cmd << std::endl; 
  
  PGresult* res = PQexec(fConnection, cmd.c_str());
  if (PQresultStatus(res) != PGRES_TUPLES_OK) {
    if (fVerbosity > 0)
      std::cerr << "SELECT failed: " << PQerrorMessage(fConnection) << std::endl;
    PQclear(res);
    return false;
  }

  // check that the number of columns is consistent
  if (PQnfields(res) != 1) {
    PQclear(res);
    return false;
  }

  // now cache rows
  int nRow = PQntuples(res);

  if (nRow != 1) {
    PQclear(res);
    return false;
  }

  if (! PQgetisnull(res,0,0)) {
    string vstr = PQgetvalue(res,0,0);
    try {
      iseq = boost::lexical_cast<long>(vstr); 
    }
    catch (boost::bad_lexical_cast &) {
      PQclear(res);
      return false;
    }
  }

  PQclear(res);

  if (!hasConn) CloseConnection();

  return true;
}

//************************************************************
bool Table::ExecuteSQL(string cmd, PGresult*& res)
{
  if (fIgnoreDB) return false;

  bool hasConn = fHasConnection;
  if (! fHasConnection) {
    GetConnection();
    hasConn = false;
  }

  if (!fConnection) {
    std::cerr << "Table::ExecuteSQL: No connection to the database!" << std::endl;
    return false;
  }

  if (cmd == "") return false;

  if (fVerbosity)
    std::cerr << "Executing SQL query: " << cmd << std::endl;

  res = PQexec(fConnection,cmd.c_str());

  // close connection to the dB if necessary
  if (! hasConn) CloseConnection();

  return (res != 0);
}

//************************************************************
bool Table::WriteToDB(bool commit)
{
  if (! CheckForNulls()) return false;

  bool doWrite = ! fIgnoreDB;
  bool hasConn = fHasConnection;

  try {
    GetConnectionInfo();
  }
  catch (std::runtime_error& e) {
    std::cerr << e.what() << std::endl;
    return false;
  }

  // make a connection to the dB if one doesn't already exist
  if (doWrite) {
    if (! fHasConnection) {
      GetConnection();
      hasConn = false;
    }  

    if (!fConnection) {
      std::cerr << "Table::WriteToDB: No connection to the database!" << std::endl;
      doWrite = false;
    }
    else {
      // now check that the table actually exists...
      if (!ExistsInDB()) {
	std::cerr << "Table::WriteToDB: Table does not exist in database!" 
		  << std::endl;
	doWrite = false;
      }
      else {
	if (! CheckValidXMLDef()) {
	  std::cerr << "Table::WriteToDB: Mismatch between XML table definition ";
	  std::cerr << "file and table defintion from the database!  ";
	  std::cerr << "Aborting write..." << std::endl;
	  doWrite = false;
	}
      }
    }
  }

  bool retVal = true;

  // get current timestamp:
  string ts = Util::GetCurrentTimeAsString();

  PGresult* res = PQexec(fConnection, "BEGIN");
  if (PQresultStatus(res) != PGRES_COMMAND_OK) {
    std::cerr << "BEGIN command failed: " << PQerrorMessage(fConnection) << std::endl;
    PQclear(res);
    CloseConnection();
    return false;
  }
  
  PQclear(res);

  std::string cmd = "SET search_path TO " + fSchema;
  res = PQexec(fConnection, cmd.c_str());
  if (PQresultStatus(res) != PGRES_COMMAND_OK) {
    std::cerr << "\'" << cmd << "\' command failed" << std::endl;
    PQclear(res);
    CloseConnection();
    return false;
  }	
  PQclear(res);
  cmd.clear();  

  // lock the table if it is of type "timestamp" or "runsubrun"; abort 
  // write if this fails
  /*
  if (TableType() == "timestamp" || 
      TableType() == "runsubrun") {
    if (fVerbosity>0)
      std::cerr << "Table::WriteToDB: locking " << Schema() << "." << Name() 
		<< " of type " << TableType() << std::endl;
    std::string cmd = "LOCK " + Schema() + "." + Name();
    res = PQexec(fConnection, cmd.c_str());
    bool isOk = true;
    if (PQresultStatus(res) != PGRES_COMMAND_OK) {
      std::cerr << "LOCK command failed: " << PQerrorMessage(fConnection) 
		<< "  Aborting write." << std::endl;
      isOk = false;
    }
    PQclear(res);
    if (!isOk) {
      CloseConnection();
      return false;
    }
  }
  */
  
  // now create the INSERT command
  for (unsigned int i=0; i<fRow.size(); ++i) {
    // do an INSERT only if this entry does not already exists in the dB
    if (! fRow[i].InDB()) {
      Row r(fRow[i]);
      if (addInsertTime) r.Set("inserttime",ts);
      if (addInsertUser) r.Set("insertuser",fUser);
      
      int nrowInsert = fCol.size();
      for (unsigned int j=0; j<fCol.size(); ++j) {
	if (r.Col(j).Name() == "updatetime") 
	  nrowInsert--;
	else if (r.Col(j).Name() == "updateuser") 
	  nrowInsert--;
	//	else if (! r.Col(j).hasValue()) 
	//	  nrowInsert--; 	  
	else if (r.Col(j).Type() == "autoincr") 
	  nrowInsert--; 	  
      }
      
      ostringstream outs;
      
      int ic=0;
      outs << "INSERT INTO " << Schema() << "." << Name() << " (";
      for (unsigned int j=0; j<fCol.size(); ++j) {
	if (r.Col(j).Name() == "updatetime") continue;
	if (r.Col(j).Name() == "updateuser") continue;
	//	if (! r.Col(j).hasValue()) continue;
	if (r.Col(j).Type() == "autoincr") continue;
	
	outs << r.Col(j).Name();
	if (ic < nrowInsert-1) outs << ",";
	++ic;
      }
      outs << ") VALUES (";
      
      ic = 0;
      for (unsigned int j=0; j<fCol.size(); ++j) {	
	if (r.Col(j).Name() == "updatetime") continue;
	if (r.Col(j).Name() == "updateuser") continue;
	//	if (! r.Col(j).hasValue()) continue;
	if (r.Col(j).Type() == "autoincr") continue;
	
	outs << r.Col(j);	
	
	if (ic < nrowInsert-1)  outs << ",";
	++ic;
      }
      
      outs << ")";
      
      if (fVerbosity > 0)
	std::cerr << "Table::WriteToDB: Executing PGSQL command: \n\t" 
		  << outs.str() << std::endl; 
      
      if (!commit)
	std::cout << outs.str() << std::endl;
      else {
	if (doWrite) {
	  res = PQexec(fConnection, outs.str().c_str());
	  if (PQresultStatus(res) != PGRES_COMMAND_OK) {
	    CacheDBCommand(outs.str());
	    std::cerr << "INSERT failed: " << PQerrorMessage(fConnection) 
		      << std::endl;
	    retVal = false;
	  }
	  else {
	    fRow[i].SetInDB();
	    // set insert columns
	    if (addInsertTime) fRow[i].Col("inserttime")->Set(ts);
	    if (addInsertUser) fRow[i].Col("insertuser")->Set(fUser);
	    // set autoincr columns
	    int64_t iseq;
	    string seqstr; 
	    for (unsigned int j=0; j<fCol.size(); ++j) {
	      if (fCol[j].Type() == "autoincr") {
		if (this->GetCurrSeqVal(fCol[j].Name(),iseq)) {
		  seqstr = boost::lexical_cast<string>(iseq);
		  fRow[i].Col(j).Set(seqstr,true);
		}
	      }
	    }
	  }
	  PQclear(res);
	}
	else 
	  CacheDBCommand(outs.str());
      }
    }
    else {
      if ( fRow[i].NModified() > 0 ) {
	Row r(fRow[i]);
	if (addUpdateTime) r.Update("updatetime",ts);
	if (addUpdateUser) r.Update("updateuser",fUser);
	std::ostringstream outs;
	outs << "UPDATE " << Schema() << "." << Name() << " SET ";
	int im = 0;
	for (unsigned int j=0; j<fCol.size() && im < r.NModified(); ++j) {
	  if (r.Col(j).Modified()) {
	    outs << r.Col(j).Name() + "=";
	    outs << r.Col(j);
	    ++im; 
	    if (im < r.NModified()) outs << ",";
	  }
	}	
	outs << " WHERE ";
	// now print out all pkey values
	int nkey = fPKeyList.size();
	for (int j=0; j<nkey; ++j) {
	  string pkey = fPKeyList[j]->Name();	    
	  outs << pkey << "=" << *r.Col(pkey);
	  if (j < (nkey-1)) outs << " and ";
	}
	
	if (fVerbosity > 0)
	  std::cerr << "Table::WriteToDB: Executing PGSQL command: \n\t" 
		    << outs.str() << std::endl; 
	
	if (!commit)
	  std::cout << outs.str() << std::endl;
	else {
	  if (doWrite) {
	    res = PQexec(fConnection, outs.str().c_str());
	    if (PQresultStatus(res) != PGRES_COMMAND_OK) {
	      CacheDBCommand(outs.str());
	      std::cerr << "UPDATE failed: " << PQerrorMessage(fConnection) << std::endl;
	      retVal = false;
	    }
	    else {
	      // set update columns
	      if (addUpdateTime) fRow[i].Col("updatetime")->Set(ts);
	      if (addUpdateUser) fRow[i].Col("updateuser")->Set(fUser);
	    }
	    PQclear(res);
	  }
	  else
	    CacheDBCommand(outs.str());
	}
      }
    }
  }

  //  PQclear(res);
  res = PQexec(fConnection, "END");
  PQclear(res);

  // close connection to the dB if necessary
  if (! hasConn) CloseConnection();

  return retVal;
}

//************************************************************
/**
 * Inserts the new Rows in this Table into the database.
 *
 * Please Note that this method does *not* provide the same level
 * of functionality as WriteToDB().  That method handles updates
 * as well as inserts, whereas this method only handles inserts.
 * Also, this method does not update the Table Rows in local memory
 * with the the correct ID (and other fields) after the Rows are
 * stored in the database.
 *
 * The expected advantage of this method is faster execution than
 * WriteToDB() for the same number of new Rows.  However, users
 * should be careful when using this method (maybe call
 * table->Clear() after calling this method to reduce the danger
 * that the Table Rows will be accessed when they have data that
 * does not fully match what is in the database).
 */
bool Table::BulkInsertInDB(int rowsPerGroup)
{
  if (! CheckForNulls()) return false;

  GetConnectionInfo();

  bool doWrite = ! fIgnoreDB;
  bool hasConn = fHasConnection;

  // make a connection to the dB if one doesn't already exist
  if (doWrite) {
    if (! fHasConnection) {
      GetConnection();
      hasConn = false;
    }

    if (!fConnection) {
      std::cerr << "Table::BulkWriteToDB: No connection to the database!" << std::endl;
      doWrite = false;
    }
    else {
      // now check that the table actually exists...
      if (!ExistsInDB()) {
        std::cerr << "Table::BulkWriteToDB: Table does not exist in database!" << std::endl;
        doWrite = false;
      }
      else {
        if (! CheckValidXMLDef()) {
          std::cerr << "Table::BulkWriteToDB: Mismatch between XML table definition file ";
          std::cerr << "and table defintion from the database!  Aborting write...";
          doWrite = false;
        }
      }
    }
  }

  bool retVal = true;

  // get current timestamp:
  string ts = Util::GetCurrentTimeAsString();

  // now create the INSERT command
  int bulkInsertCount = 0;
  ostringstream outs;
  outs.str("");
  for (unsigned int i=0; i<fRow.size(); ++i) {
    // do an INSERT only if this entry does not already exists in the dB
    if (! fRow[i].InDB()) {
      Row r(fRow[i]);
      if (addInsertTime) r.Set("inserttime",ts);
      if (addInsertUser) r.Set("insertuser",fUser);

      int nrowInsert = fCol.size();
      for (unsigned int j=0; j<fCol.size(); ++j) {
        if (r.Col(j).Name() == "updatetime")
          nrowInsert--;
        else if (r.Col(j).Name() == "updateuser")
          nrowInsert--;
        else if (r.Col(j).Type() == "autoincr")
          nrowInsert--;
      }

      int ic;
      if (bulkInsertCount == 0) {
        ic = 0;
        outs << "INSERT INTO " << Schema() << "." << Name() << " (";
        for (unsigned int j=0; j<fCol.size(); ++j) {
          if (r.Col(j).Name() == "updatetime") continue;
          if (r.Col(j).Name() == "updateuser") continue;
          if (r.Col(j).Type() == "autoincr") continue;

          outs << r.Col(j).Name();
          if (ic < nrowInsert-1) outs << ",";
          ++ic;
        }
        outs << ") VALUES (";
      }
      else {
        outs << ", (";
      }

      ic = 0;
      for (unsigned int j=0; j<fCol.size(); ++j) {
        if (r.Col(j).Name() == "updatetime") continue;
        if (r.Col(j).Name() == "updateuser") continue;
        if (r.Col(j).Type() == "autoincr") continue;

	outs << r.Col(j);
	
        if (ic < nrowInsert-1)  outs << ",";
        ++ic;
      }

      outs << ")";

      ++bulkInsertCount;
      if (bulkInsertCount >= rowsPerGroup) {
        if (fVerbosity > 0)
          std::cerr << "Table::BulkWriteToDB: Executing PGSQL command: \n\t"
		    << outs.str() << std::endl;

        if (doWrite) {
          PGresult* res = PQexec(fConnection, outs.str().c_str());
          if (PQresultStatus(res) != PGRES_COMMAND_OK) {
            CacheDBCommand(outs.str());
            std::cerr << "INSERT failed: " << PQerrorMessage(fConnection) << std::endl;
            retVal = false;
          }
          PQclear(res);
        }
        else {
          CacheDBCommand(outs.str());
        }

        bulkInsertCount = 0;
        outs.str("");
      }
    }
  }

  if (bulkInsertCount > 0) {
    if (fVerbosity > 0)
      std::cerr << "Table::WriteToDB: Executing PGSQL command: \n\t"
		<< outs.str() << std::endl;

    if (doWrite) {
      PGresult* res = PQexec(fConnection, outs.str().c_str());
      if (PQresultStatus(res) != PGRES_COMMAND_OK) {
        CacheDBCommand(outs.str());
        std::cerr << "INSERT failed: " << PQerrorMessage(fConnection) << std::endl;
        retVal = false;
      }
      PQclear(res);
    }
    else {
      CacheDBCommand(outs.str());
    }
  }
  
  // close connection to the dB if necessary
  if (! hasConn) CloseConnection();

  return retVal;
}

//************************************************************
bool Table::WriteToCSV(string fname, bool appendToFile)
{

  if (! CheckForNulls()) return false;

  ofstream fout;
  if (!appendToFile)
    fout.open(fname.c_str());
  else
    fout.open(fname.c_str(),ios_base::app);

  for (unsigned int i=0; i<fRow.size(); ++i) {
    for (unsigned int j=0; j<fCol.size(); ++j) {
      fout << fRow[i].Col(j);
      if (j<fCol.size()-1) fout << ",";
    }
    fout << std::endl;
  }

  fout.close();
  
  return true;
}

//************************************************************
void Table::RemoveValidityRange(std::string& cname)
{  
  unsigned int i=0;
  for (; i<fValidityStart.size(); ++i) 
    if (fValidityStart[i].Name() == cname) {
      fValidityStart.erase(fValidityStart.begin()+i);
      fValidityEnd.erase(fValidityEnd.begin()+i);
    }
}

//************************************************************
// Add the ith column in the table to the list of columns
// that are to be "distinct"
//************************************************************
bool Table::AddDistinctColumn(unsigned int i)
{  
  if (i >= fCol.size()) return false;

  const Column* c = &fCol[i];
  for (unsigned int j=0; j<fDistinctCol.size(); ++j)
    if (fDistinctCol[j] == c) return true;
  
  fDistinctCol.push_back(c);
    
  return true;
}

//************************************************************
// Remove the ith column in the table to the list of columns
// that are to be "distinct"
//************************************************************
bool Table::RemoveDistinctColumn(unsigned int i)
{  
  if (i >= fCol.size()) return false;

  const Column* c = &fCol[i];
  for (unsigned int j=0; j<fDistinctCol.size(); ++j)
    if (fDistinctCol[j] == c) {
      fDistinctCol.erase(fDistinctCol.begin() + j);
      return true;
    }
  
  return false;
}

//************************************************************
// Add the column with name "cname" in the table to the list 
// of columns that are to be "distinct"
//************************************************************
bool Table::AddDistinctColumn(string cname)
{  
  const Column* c = GetCol(cname);
  
  if (!c) return false;

  for (unsigned int j=0; j<fDistinctCol.size(); ++j)
    if (fDistinctCol[j] == c) return true;
  
  fDistinctCol.push_back(c);
    
  return true;
}

//************************************************************
// Remove the column with name "cname" in the table to the 
// list of columns that are to be "distinct"
//************************************************************
bool Table::RemoveDistinctColumn(string cname)
{  
  const Column* c = GetCol(cname);
  
  if (!c) return false;

  for (unsigned int j=0; j<fDistinctCol.size(); ++j)
    if (fDistinctCol[j] == c) {
      fDistinctCol.erase(fDistinctCol.begin() + j);
      return true;
    }
  
  return false;
}

//************************************************************
// Add the ith column in the table to the list of columns
// that define the sorting order
//************************************************************
bool Table::AddOrderColumn(unsigned int i)
{  
  if (i >= fCol.size()) return false;

  const Column* c = &fCol[i];
  for (unsigned int j=0; j<fOrderCol.size(); ++j)
    if (fOrderCol[j] == c) return true;
  
  fOrderCol.push_back(c);
    
  return true;
}

//************************************************************
// Remove the ith column in the table to the list of columns
// that define the sorting order
//************************************************************
bool Table::RemoveOrderColumn(unsigned int i)
{  
  if (i >= fCol.size()) return false;

  const Column* c = &fCol[i];
  for (unsigned int j=0; j<fOrderCol.size(); ++j)
    if (fOrderCol[j] == c) {
      fOrderCol.erase(fOrderCol.begin() + j);
      return true;
    }
  
  return false;
}

//************************************************************
// Add the column with name "cname" in the table to the list 
// of columns that define the sorting order
//************************************************************
bool Table::AddOrderColumn(string cname)
{  
  const Column* c = GetCol(cname);
  
  if (!c) return false;

  for (unsigned int j=0; j<fOrderCol.size(); ++j)
    if (fOrderCol[j] == c) return true;
  
  fOrderCol.push_back(c);
    
  return true;
}

//************************************************************
// Remove the column with name "cname" in the table to the 
// list of columns that define the sorting order
//************************************************************
bool Table::RemoveOrderColumn(string cname)
{  
  const Column* c = GetCol(cname);
  
  if (!c) return false;

  for (unsigned int j=0; j<fOrderCol.size(); ++j)
    if (fOrderCol[j] == c) {
      fOrderCol.erase(fOrderCol.begin() + j);
      return true;
    }
  
  return false;
}


