//  This file (ShmRdWr.cpp) was created by Ron Rechenmacher <ron@fnal.gov> on
//  May 23, 2011. "TERMS AND CONDITIONS" governing this file are in the README
//  or COPYING file. If you do not have such a file, one can be obtained by
//  contacting Ron or Fermi Lab in Batavia IL, 60510, phone: 630-840-3000.
//  $RCSfile: ShmRdWr.cpp,v $
//  rev="$Revision: 1.30 $$Date: 2015/07/16 14:55:35 $";

/* The writer will do (in the following order):
   1) inc idx_wr_start
   1a) determine buffer
   2) write time in buffer
   3) write byte count and data in buffer
   4) inc total_bytes_written
   5) inc idxcnt_wr_complete

Readers init their own static read_idx
   1) readers checks it's read_idx against the idxcnt_wr_complete to determine
      how many buffers are available and if any have been missed (overwritten
      since last read.
   2) if buffers are available, read a buffer
   3) after a buffer has been read, check cnt associatedd with that buffer
      against idx_wr_start to see if buffer just read has been written to
      (at least _potentially)
*/

#include <stdio.h>		// printf
#include <stdlib.h>		// getenv
#include <sys/shm.h>		// shmget
#include <sys/sem.h>		// semget
#include <errno.h>		// errno, EACCES, EINVAL
#include <strings.h>		// bzero
#include <string.h>		// memcpy
#include <sys/time.h>		// gettimeofday
#include <stdarg.h>		// va_start,...
#include "../include/ShmRdWr.h"

// make CXXFLAGS=-DDEBUG
#ifdef DEBUG
# define DPRINT(x) tprint x
#else
# define DPRINT(x)
#endif
#ifdef TEST
# define PAUSE( msg )	{ char xx[99];printf(msg);printf("press enter/return to continue...");fflush(stdout);fgets(xx,sizeof(xx),stdin); }
#else
# define PAUSE( msg )
#endif



ShmRdWr::ShmRdWr(int options, shmrw_gid_t default_gIdx)
    : shm_hdr_ptr_(NULL)
    , options_(options)
    , rdr_info_((ShmRdWr_Rdr){0,0})
    , shm_grprdr_info__ptr_(NULL)
    , have_sem_(250,false)
    , gIdx_dflt_(default_gIdx)
    , poll_ms_(5)
{
}

bool const ShmRdWr::semKeepAfterDataAvail=true;

// returns: 0=success 1=error
int ShmRdWr::attach()
{
    key_t	shmkey;
    uint16_t	bufsNum;
    uint16_t	bufsSiz_MB;
    uint8_t	grpsNum;
    uint8_t	grpsSiz_pages;

    if (getShmKeyParams(shmkey,bufsNum,bufsSiz_MB,grpsNum,grpsSiz_pages))
    {   DPRINT(("ERROR - problems with getShmKeyParams\n"));
	return (1);
    }

    // Try to attach, creating if it doesn't exist.
    size_t shmsize = paramsToSize( bufsNum,bufsSiz_MB,grpsNum,grpsSiz_pages );
    DPRINT(("shmsize=%lld\n",(long long)shmsize));
    int    shmflgs =  IPC_CREAT | IPC_EXCL | 0666; // 1st try to create and error if one exists
    int shmid = shmget( shmkey, shmsize, shmflgs );
    if (shmid != -1)
    {   // NEWLY CREATED - get ptr, add semaphores
	shm_hdr_ptr_ = (ShmRdWr_Header*)shmat( shmid, NULL, 0 );
	if (shm_hdr_ptr_ == (void *)-1) return (1);
	bzero( shm_hdr_ptr_, sizeof(*shm_hdr_ptr_) );
	shm_hdr_ptr_->bufs_num      = bufsNum;
	shm_hdr_ptr_->buf_sz_MB     = bufsSiz_MB;
	shm_hdr_ptr_->buf_sz_bytes  = bufsSiz_MB*0x100000;
	shm_hdr_ptr_->grpsNum       = grpsNum;
	shm_hdr_ptr_->grpsSiz_pages = grpsSiz_pages;
	shm_hdr_ptr_->bufs_largest_multiple = 0xffffffff - (0xffffffff%bufsNum);
	shm_hdr_ptr_->largest_zero_offset = (0xffffffff%bufsNum) + 1;
	shm_hdr_ptr_->semSetId = -1;

	if (grpsNum)
	{   int semid = semget( shmkey, grpsNum, IPC_CREAT | 0666 );  // no IPC_EXCL - may be left over
	    if (semid == -1) { perror( "semget" ); return (1); }
	    // initialize the sem set (all) to 1 without the SEM_UNDO flag
	    sembuf _sembuf[SEMMSL];
	    for (int ii=0; ii<grpsNum; ++ii)
	    {   _sembuf[ii].sem_num = ii;
		_sembuf[ii].sem_op = 1;
		_sembuf[ii].sem_flg = 0;
	    }
	    if (semop(semid,_sembuf,grpsNum) == -1)
	    {   perror("semop");
		return (1);
	    }
	    shm_hdr_ptr_->semSetId = semid;
	}
    }
    else if (errno == EEXIST)
    {   // There is an exist segment -- get/attach to it
	shmflgs = 0;
	shmid =  shmget( shmkey, shmsize, shmflgs );
	if (shmid != -1)
	{   // have id of existing - attach and check for (application) consistency
	    shm_hdr_ptr_ = (ShmRdWr_Header*)shmat( shmid, NULL, 0);
	    if (shm_hdr_ptr_ == (void *)-1)
	    {   perror("shmat");
		return (1);
	    }
	    if (grpsNum && (shm_hdr_ptr_->semSetId == -1))
		sleep(1); // wait for semaphores ???
	    if (grpsNum && (shm_hdr_ptr_->semSetId == -1))
	    {   DPRINT(("ERROR - grpsNum && semSetId=-1\n"));
		return (1);
	    }
	}
	else
	{   perror( "shmget" );
	}
    }
    else
    {   // a real error
	perror( "shmget" );
	return (1);
    }
    shm_databuf_metadata_ptr_ = (ShmRdWr_BufMetadata*)(((uint8_t*)shm_hdr_ptr_) + 4096);
    shm_grprdr_info__ptr_     = (ShmRdWr_Rdr*)(((uint8_t*)shm_databuf_metadata_ptr_) + 2*4096);
    shm_prv_data_ptr_         = (uint8_t*)(((uint8_t*)shm_grprdr_info__ptr_) + 2*4096);
    shm_databuf_data_ptr_     = shm_prv_data_ptr_ + grpsNum*4096;
    if (options_ & SHMRW_O_WRITER)
    {   // set writer pid
	if (shm_hdr_ptr_->writer_pid)
	{
	}
	shm_hdr_ptr_->writer_pid = getpid();
	// cancel any write in-progress (for now, ASSUME previous writer died)
	shm_hdr_ptr_->idx_wr_start = shm_hdr_ptr_->idxcnt_wr_complete;
	shm_hdr_ptr_->offset = 0;
    }
    return (0);
}   // attach


// error returns: -28 ENOSPC
ssize_t ShmRdWr::write( void* src_ptr, size_t byte_cnt, bool more, off_t off )
{
    if ((shm_hdr_ptr_==NULL) && (attach()!=0)) return (-1);
    DPRINT(("write: off=%ld\n", off ));

    int bufidx;
    if (shm_hdr_ptr_->idx_wr_start == shm_hdr_ptr_->idxcnt_wr_complete)
    {   // starting new write
	if (byte_cnt > (size_t)shm_hdr_ptr_->buf_sz_MB*0x100000) return (-2);
	// FIRST adjust idx_wr_start
	shm_hdr_ptr_->idx_wr_start = bufcnt_add( shm_hdr_ptr_->idxcnt_wr_complete, 1 );
	shm_hdr_ptr_->offset = 0;
	bufidx = bufcnt2idx( shm_hdr_ptr_->idx_wr_start );
	bzero( &shm_databuf_metadata_ptr_[bufidx], sizeof(ShmRdWr_BufMetadata) );
    }
    else // continuing write
    {   bufidx=bufcnt2idx( shm_hdr_ptr_->idx_wr_start );
    }
	
    if (off != -1) shm_hdr_ptr_->offset = off;
    if (byte_cnt+shm_hdr_ptr_->offset > (size_t)shm_hdr_ptr_->buf_sz_MB*0x100000)
	return (-ENOSPC);

    uint8_t* datptr=shm_databuf_data_ptr_ + bufidx*shm_hdr_ptr_->buf_sz_MB*0x100000;

    DPRINT(("write: shm_hdr_ptr_->offset=%ld\n", shm_hdr_ptr_->offset ));
    memcpy( datptr+shm_hdr_ptr_->offset, src_ptr, byte_cnt );

    shm_hdr_ptr_->offset+=byte_cnt;

    if (shm_hdr_ptr_->offset > (off_t)shm_databuf_metadata_ptr_[bufidx].data_bytes)
	shm_databuf_metadata_ptr_[bufidx].data_bytes = shm_hdr_ptr_->offset;

    if (more == false)
    {   gettimeofday( &shm_databuf_metadata_ptr_[bufidx].wr_complete_tod, NULL );
	shm_hdr_ptr_->idxcnt_wr_complete = shm_hdr_ptr_->idx_wr_start;
	// LAST - adjust idxcnt_wr_complete
    }
    return (byte_cnt);
}
ssize_t ShmRdWr::write( void* src_ptr, size_t byte_cnt, off_t off, bool more )
{   return write( src_ptr, byte_cnt, more, off );
}
ssize_t ShmRdWr::write( void* src_ptr, size_t byte_cnt )
{   return write( src_ptr, byte_cnt, false, (off_t)-1 );
}   // write


// Get the next buffer to be read and Info, but does not "start" read (incase calling method has (allowPartial=false && obufSz<bytesInBuf))
// returns bufIdx (>=0) OR error/tmo (<0)
int ShmRdWr::read_next_( ShmRdWr_Info& info, timeval& expire, ShmRdWr_Rdr& rdr )
{
    if ((shm_hdr_ptr_==NULL) && (attach()!=0)) return (-1);
    int bufidx=-1;

    bzero( &info, sizeof(info) ); // init info

    uint32_t buffs_num=shm_hdr_ptr_->bufs_num;


    // check if already started
    if (rdr.idx_rd_start != rdr.idxcnt_rd_complete)
    {   bufidx = bufcnt2idx( rdr.idx_rd_start );
	info.bufInfo  = shm_databuf_metadata_ptr_[bufidx];
	info.data_ptr = shm_databuf_data_ptr_+ bufidx*shm_hdr_ptr_->buf_sz_MB*0x100000;
	// check if bufInfo is OK
	if (bufcnt_delta(shm_hdr_ptr_->idx_wr_start,rdr.idx_rd_start) < buffs_num)
	{   // yes, it is OK (at this point)
	    return (bufcnt2idx( rdr.idx_rd_start ));
	}
	// else not OK -- end read, return error
	++rdr.total_missed;
	rdr.idxcnt_rd_complete = rdr.idx_rd_start;
	rdr.offset = 0;
	++info.ovrwt;
	info.ovrwt_tot = rdr.total_missed;
	return (-SHMRW_EOVRWT);
    }

    while (1)
    {
	uint32_t buffs_delta = bufcnt_delta( shm_hdr_ptr_->idxcnt_wr_complete
					    ,rdr.idxcnt_rd_complete );
	if (buffs_delta > buffs_num)
	{   // missed some
	    uint32_t missed = buffs_delta - buffs_num;
	    rdr.total_missed += missed;
	    // reset idxcnt_rd_complete to recover
	    rdr.idxcnt_rd_complete = bufcnt_add( shm_hdr_ptr_->idx_wr_start, -buffs_num );
	    if (buffs_num > 1)
	    {   // sacrifice a buffer if we can to help assure recovery
		++rdr.idxcnt_rd_complete;
		++missed;
		++rdr.total_missed;
	    }
	    info.ovrwt += missed;
	    // now go on to actually read a buffer (below)...
	}
	else if (buffs_delta == 0)
	{   // no buffers available
	    if ((expire.tv_sec!=0) && is_tmo(expire))
	    {   info.ovrwt_tot = rdr.total_missed;
		info.cushion = buffs_num - ((  shm_hdr_ptr_->idx_wr_start
					     !=shm_hdr_ptr_->idxcnt_wr_complete)?1:0);
		DPRINT(("read_next_: timeout\n"));
		return (-EAGAIN);
	    }
	    DPRINT(("read_next_: poll sleep %ums\n",poll_ms_));
	    usleep( poll_ms_*1000 );
	    continue;
	}
	// else 0 < buffs_delta <= buffs_num

	// read next buffer
	rdr.idx_rd_start = bufcnt_add( rdr.idxcnt_rd_complete, 1 );
	rdr.offset = 0;
	bufidx = bufcnt2idx( rdr.idx_rd_start );
	info.bufInfo  = shm_databuf_metadata_ptr_[bufidx];
	info.data_ptr = shm_databuf_data_ptr_+ bufidx*shm_hdr_ptr_->buf_sz_MB*0x100000;
	info.ovrwt_tot = rdr.total_missed;
	// check if bufInfo is OK
	if (bufcnt_delta(shm_hdr_ptr_->idx_wr_start,rdr.idx_rd_start) < buffs_num)
	{   // yes, it is OK (at this point)
	    break;
	}
    }
    return (bufidx);
}   // read_next_

// returns: >=0 data bytes read
//          -4 in prog ovrwt  SHM_EOVRWT
//          -28 obuf too small ENOSPC   (only when allowPartial=false)
//          -11 timeout       EAGAIN
ssize_t ShmRdWr::read_(  void* dst_ptr,      size_t dst_max_bytes, ShmRdWr_Info& info
		     , timeval& expire_time, bool allowPartial, ShmRdWr_Rdr& rdr )
{
    if ((shm_hdr_ptr_==NULL) && (attach()!=0)) return (-1);
    int	retsts=0;
#   if 0
    if (&info != NULL) printf("&info=%p\n", &info );
    else               printf("&info=NULL\n" );
#   endif

    bzero( &info, sizeof(info) ); // init info

    uint32_t buffs_num=shm_hdr_ptr_->bufs_num;

    while (1)
    {   int bufidx;
	if ((bufidx=read_next_(info,expire_time,rdr)) < 0)
	{   retsts=bufidx;
	    break;
	}

	size_t	 datbytes  = shm_databuf_metadata_ptr_[bufidx].data_bytes;
	uint8_t* datptr    = shm_databuf_data_ptr_ + bufidx*shm_hdr_ptr_->buf_sz_MB*0x100000;

	if (!allowPartial && ((datbytes-rdr.offset)>dst_max_bytes))
	{   retsts = -ENOSPC;  // didn't read b/c obuf too small
	    break;
	}
	else
	{   size_t memcpy_bytes=((datbytes-rdr.offset)>dst_max_bytes)
		? dst_max_bytes: (datbytes-rdr.offset);
	    DPRINT(("read_: memcpy(dest,datptr+%ld,%ld)\n",rdr.offset,memcpy_bytes));
	    memcpy( dst_ptr, datptr+rdr.offset, memcpy_bytes );
	    rdr.offset += memcpy_bytes;
	    // check for possible/probable overwrite
	    if (bufcnt_delta(shm_hdr_ptr_->idx_wr_start,rdr.idx_rd_start) >= buffs_num)
	    {   // A PROBABLE OVRWT - start OVER
		continue;
	    }
	    else
	    {   // ALL IS WELL
		// inc idxcnt, fill in info and return
		if (!allowPartial || (rdr.offset==(unsigned)datbytes))
		    rdr.idxcnt_rd_complete = rdr.idx_rd_start;
		retsts = memcpy_bytes;
		info.ovrwt_tot = rdr.total_missed;
		info.cushion = buffs_num - ((  shm_hdr_ptr_->idx_wr_start
					     !=shm_hdr_ptr_->idxcnt_wr_complete)?1:0);
		info.bufInfo = shm_databuf_metadata_ptr_[bufidx];
		break;
	    }
	}
    }   // while (1)

    return (retsts);
}   // read_


ssize_t ShmRdWr::read( void* dst_ptr,     size_t dst_max_bytes
		    , ShmRdWr_Info& info, time_t tmo_ms,       bool allowPartial )
{   timeval expire;
    return read_( dst_ptr, dst_max_bytes, info, set_expire_time(expire,tmo_ms), allowPartial, rdr_info_ );
}
ssize_t ShmRdWr::read( void* dst_ptr, size_t max_bytes, uint32_t& ovrwt, time_t tmo_ms )
{   ShmRdWr_Info info;
    int sts=read( dst_ptr, max_bytes, info, tmo_ms );
    if (&ovrwt!= NULL)
    {   ovrwt = info.ovrwt;
    }
    return (sts);
}
ssize_t ShmRdWr::read( void* dst_ptr, size_t max_bytes, time_t tmo_ms, uint32_t& ovrwt )
{   return read( dst_ptr, max_bytes, ovrwt, tmo_ms );
}   // read



ssize_t ShmRdWr::read_grp(  void*  dptr, size_t dbytes,               ShmRdWr_Info& info
			  , time_t tmo,  bool   semKeepAfterDataAvail, shmrw_gid_t gIdx )
{
    if ((shm_hdr_ptr_==NULL) && (attach()!=0)) return (-1);
    DPRINT(("read_grp(...dbytes=%ld,info,gIdx=%u,tmo=%ld)\n",dbytes,(unsigned)gIdx,tmo));
    if (gIdx >= SEMMSL) gIdx = gIdx_dflt_;
    if (gIdx >= shm_hdr_ptr_->grpsNum) return (-1);

    if (have_sem_[gIdx] && (tmo!=0))
    {   DPRINT(("read_grp: setting tmo=0\n"));
	tmo=0;
    }
    timeval	expire_time;
    set_expire_time( expire_time, tmo );

    int sts=0;
    while (1)
    {
	if (have_sem_[gIdx] == false)
	{   // get sem, remember to do release
	    timespec tmo_ts;
	    expire_tv2ts( expire_time, tmo_ts );
	    if (!semtake(gIdx,tmo_ts)) return (0);
	    have_sem_[gIdx] = true;
	}

	// all read_ with tmo=0 as we are handling any tmo!=0 at this level
	timeval tmo0((timeval){1,0}); // a long time ago
	sts = read_( dptr, dbytes, info, tmo0, semKeepAfterDataAvail, shm_grprdr_info__ptr_[gIdx] );
	DPRINT(("read_grp: read_=%d\n",sts));
	if (sts != -EAGAIN)
	    break;
	if ((tmo!=-1) && is_tmo(expire_time))
	{   DPRINT(("read_grp: timed out\n"));
	    break;
	}
	semgive( gIdx );
	have_sem_[gIdx] = false;
	DPRINT(("read_grp: poll sleep %ums exp=*.%06ld\n",poll_ms_,expire_time.tv_usec));
	usleep( poll_ms_*1000 );
    }
    //  Any error should cause semgive
    if (  (have_sem_[gIdx]==true)
	&&(!(semKeepAfterDataAvail&&((sts>=0)&&info.bufInfo.data_bytes))) )
    {   semgive( gIdx );
	have_sem_[gIdx] = false;
    }
    return (sts);
}
ssize_t ShmRdWr::read_grp(  void*dest,  size_t dbytes,               uint32_t&   ovrwt
			, time_t tmo, bool   semKeepAfterDataAvail, shmrw_gid_t gIdx )
{   ShmRdWr_Info info;
    int sts=read_grp( dest, dbytes, info, tmo, semKeepAfterDataAvail, gIdx );
    if (&ovrwt!= NULL)
    {   ovrwt = info.ovrwt;
    }
    return (sts);
}
ssize_t ShmRdWr::read_grp(  void*dest,  size_t max_bytes
			, time_t tmo, bool   semKeepAfterDataAvail, shmrw_gid_t gIdx )
{   //ShmRdWr_Info info;
    return read_grp( dest, max_bytes, *(ShmRdWr_Info*)0, tmo, false, gIdx );
}   // read_grp

ssize_t ShmRdWr::read_grp_start( ShmRdWr_Info&info, time_t tmo,shmrw_gid_t gIdx )
{   ssize_t sts=read_grp(0,0,info,tmo,true,gIdx);
    if (sts<0) return (sts);
    else
    {   // return bytes avail instead of bytes read (which would be 0,as specified)
	return(info.bufInfo.data_bytes);
    }
}
ssize_t ShmRdWr::read_grp_start( uint32_t&ovrwt, time_t tmo,shmrw_gid_t gIdx )
{   ShmRdWr_Info info;ssize_t sts=read_grp_start(info,tmo,gIdx);
    ovrwt = info.ovrwt;
    return (sts);
}

int ShmRdWr::read_grp_check( uint32_t&ovrwt, shmrw_gid_t gIdx )
{
    int retsts=0;
    if (gIdx >= SEMMSL) gIdx = gIdx_dflt_;
    if (gIdx >= shm_hdr_ptr_->grpsNum) return (-1);
    ShmRdWr_Rdr& rdr=shm_grprdr_info__ptr_[gIdx];
    DPRINT(("read_grp_close: gIdx=%u start=%u complete=%u\n",gIdx,rdr.idx_rd_start, rdr.idxcnt_rd_complete));
    if (rdr.idx_rd_start != rdr.idxcnt_rd_complete)
    {   // check for possible ovrwt
	if (bufcnt_delta(shm_hdr_ptr_->idx_wr_start,rdr.idx_rd_start) >= shm_hdr_ptr_->bufs_num)
	{   // ovrwt
	    ovrwt=1;
	    ++rdr.total_missed;
	    // close and return indication
	    rdr.idxcnt_rd_complete = rdr.idx_rd_start;
	    rdr.offset = 0;
	    retsts = -SHMRW_EOVRWT;
	}
	else ovrwt=0;
    }
    else // already closed -- no additional ovrwt
	ovrwt=0;
    return (retsts);
}   // read_grp_check

int ShmRdWr::read_grp_close( uint32_t&ovrwt, bool semKeepAfterDataAvail, shmrw_gid_t gIdx )
{
    if ((shm_hdr_ptr_==NULL) && (attach()!=0)) return (-1);
    if (gIdx >= SEMMSL) gIdx = gIdx_dflt_;
    if (gIdx >= shm_hdr_ptr_->grpsNum) return (-1);
    ShmRdWr_Rdr& rdr=shm_grprdr_info__ptr_[gIdx];
    DPRINT(("read_grp_close: gIdx=%u start=%u complete=%u\n",gIdx,rdr.idx_rd_start, rdr.idxcnt_rd_complete));
    if (rdr.idx_rd_start != rdr.idxcnt_rd_complete)
    {   // need to close; check for possible ovrwt
	if (bufcnt_delta(shm_hdr_ptr_->idx_wr_start,rdr.idx_rd_start) >= shm_hdr_ptr_->bufs_num)
	{   // ovrwt
	    ovrwt=1;
	    ++rdr.total_missed;
	}
	else ovrwt=0;
	rdr.idxcnt_rd_complete = rdr.idx_rd_start;
	rdr.offset = 0;
    }
    else // already closed -- no additional ovrwt
	ovrwt=0;
    if (  (have_sem_[gIdx]==true) && !(semKeepAfterDataAvail) )
    {   semgive( gIdx );
	have_sem_[gIdx] = false;
    }
    return (0);
}   // read_grp_close


// compatibility method
int ShmRdWr::start_read( ShmRdWr_BufMetadata**hdr, ShmRdWrCnt& missed,  uint32_t & cushion, int tmo_us )
{   int retsts=0;
    ShmRdWr_Info info;
    ssize_t ss=read_grp_start( info, tmo_us/1000 );
    missed = info.ovrwt;
    if (ss == 0) // 0 data is "no data"
    {   uint32_t ovrwt;
	read_grp_close( ovrwt );
	missed += ovrwt;
	retsts = -1;
    }
    else
    {   *hdr = (ShmRdWr_BufMetadata*)info.data_ptr; // regardless of whether ss<0
	retsts = ss;
    }
    return (retsts);
}

int ShmRdWr::read(  void*     dstbuf,       size_t      dstbuf_sz_bytes
		  , timeval&  this_or_last, ShmRdWrCnt& missed
		  , uint32_t& cushion,      uint32_t&   more_avail, int tmo_us )
{   ShmRdWr_Info info;
    time_t tmo_ms;
    if (tmo_us == -1) tmo_ms = -1;
    else              tmo_ms = tmo_us/1000;
    ssize_t ss=read_grp( dstbuf, dstbuf_sz_bytes, info, tmo_ms );
    this_or_last = info.bufInfo.wr_complete_tod;
    missed = info.ovrwt;
    cushion = info.cushion;
    return (ss);
}


int ShmRdWr::read_grp_prv(  void* dst_ptr, size_t max_bytes
			, time_t tmo, bool semKeep, shmrw_gid_t grpIdx )
{

    return (0);
}   // read_grp_prv

int ShmRdWr::write_grp_prv(  void* src_ptr, size_t bytes_to_write
			 , time_t tmo,    bool   semKeep,    shmrw_gid_t grpIdx )
{
    
    return (0);
}   // write_grp_prv

int ShmRdWr::get_grp_missed( uint32_t &missed, uint32_t &new_missed, shmrw_gid_t gIdx )
{
    if ((shm_hdr_ptr_==NULL) && (attach()!=0)) return (-1);
    if (gIdx >= SEMMSL) gIdx = gIdx_dflt_;
    if (gIdx >= shm_hdr_ptr_->grpsNum) return (-1);
    ShmRdWr_Rdr& rdr=shm_grprdr_info__ptr_[gIdx];

    missed = rdr.total_missed;
    new_missed = rdr.total_missed
	+ bufcnt_delta(shm_hdr_ptr_->idx_wr_start,rdr.idx_rd_start);

    return (0);
}   // get_grp_missed


int ShmRdWr::info(  size_t   & buf_payload,    size_t   & buf_sz_bytes
		  , uint32_t & poll_period_us, uint32_t & version
		  , pid_t    & writer_pid )
{
    return (0);
}

uint64_t ShmRdWr::get_us_timeofday()
{   timeval tv;
    gettimeofday( &tv, NULL );
    return (uint64_t)tv.tv_sec*1000000+tv.tv_usec;
}   // get_us_timeofday

int ShmRdWr::print_header( int fd )
{
    return (0);
}

std::ostream& operator <<( std::ostream& out, ShmRdWr& shmrw )
{
    char bigbuf[4096];
    snprintf( bigbuf, sizeof(bigbuf), "\
                     bufs_num=%u\n\
                    buf_sz_MB=%u\n\
                      grpsNum=%u\n\
                grpsSiz_pages=%u\n\
        bufs_largest_multiple=0x%08x\n\
          largest_zero_offset=0x%08x\n\
                 idx_wr_start=%u\n\
           idxcnt_wr_complete=%u\n\
                       offset=%ld\n\
          total_bytes_written=%u\n\
                     semSetId=%d\n\
                   writer_pid=%u\n\
"
		   , shmrw.shm_hdr_ptr_->bufs_num
		   , shmrw.shm_hdr_ptr_->buf_sz_MB
		   , shmrw.shm_hdr_ptr_->grpsNum
		   , shmrw.shm_hdr_ptr_->grpsSiz_pages
		   , shmrw.shm_hdr_ptr_->bufs_largest_multiple
		   , shmrw.shm_hdr_ptr_->largest_zero_offset
		   , shmrw.shm_hdr_ptr_->idx_wr_start
		   , shmrw.shm_hdr_ptr_->idxcnt_wr_complete
		   , shmrw.shm_hdr_ptr_->offset
		   , shmrw.shm_hdr_ptr_->total_bytes_written
		   , shmrw.shm_hdr_ptr_->semSetId
		   , shmrw.shm_hdr_ptr_->writer_pid
		   );
    std::string ostr(bigbuf);
    snprintf( bigbuf, sizeof(bigbuf), "idx  _bytes_  _1st_lword\n" );
    ostr += bigbuf;
    for (int ii=0; ii<shmrw.shm_hdr_ptr_->bufs_num; ++ii)
    {   uint32_t *dptr=(uint32_t*)(  shmrw.shm_databuf_data_ptr_
				   + ii*shmrw.shm_hdr_ptr_->buf_sz_MB*0x100000);
	snprintf( bigbuf, sizeof(bigbuf), "%3d  %7ld ",ii
		 ,shmrw.shm_databuf_metadata_ptr_[ii].data_bytes );
	ostr += bigbuf;
	unsigned jj;
	for (jj=0; jj<6; ++jj)
	{   snprintf( bigbuf, sizeof(bigbuf), " 0x%08x", dptr[jj] );
	    ostr += bigbuf;
	}
	ostr+="\n             ";
	for (; jj<12; ++jj)
	{   snprintf( bigbuf, sizeof(bigbuf), " 0x%08x", dptr[jj] );
	    ostr += bigbuf;
	}
	ostr+='\n';
    }
    for (int ii=0; ii<shmrw.shm_hdr_ptr_->grpsNum; ++ii)
    {   int semval = semctl( shmrw.shm_hdr_ptr_->semSetId, ii, GETVAL );
	uint32_t missed, new_missed;
	shmrw.get_grp_missed( missed, new_missed, ii );
	snprintf( bigbuf, sizeof(bigbuf), "\
grpIdx=%3d idxcnt_rd_complete=%u\n\
                 idx_rd_start=%u\n\
                 total_missed=%u  %u\n\
                       offset=%ld\n\
                       semval=%d\n\
"
		 , ii
		 , shmrw.shm_grprdr_info__ptr_[ii].idxcnt_rd_complete
		 , shmrw.shm_grprdr_info__ptr_[ii].idx_rd_start
		 , missed, new_missed
		 , shmrw.shm_grprdr_info__ptr_[ii].offset
		 , semval
		 );
	ostr += bigbuf;
    }
    out << ostr;
    return (out);
}   // operator <<


key_t ShmRdWr::getShmKey()
{   key_t retkey;
    const char *key=envvar();
    const char *envval=getenv(key);
    if (envval) retkey=strtoul(envval,NULL,0);
    else        retkey=SHMRW_DEFAULT_KEY;
    return (retkey);
}   // getShmKey


// returns: 0=success 1=error
int ShmRdWr::getShmKeyParams(  key_t&    shmkey
			   , uint16_t& bufsNum, uint16_t& bufsSiz_MB
			   , uint8_t&  grpsNum, uint8_t&  grpsSiz_pages )
{
    shmkey        = getShmKey();
    bufsNum       =  (shmkey      & 0xff) + 1;
    bufsSiz_MB    = ((shmkey>>8)  & 0xff) + 1;
    grpsNum       = ((shmkey>>16) & 0xff);     // 0 is valid
    grpsSiz_pages = ((shmkey>>24) & 0xf)  + 1;
    return (0);
}   // getShmKeyParams

size_t ShmRdWr::paramsToSize(  uint16_t bufsNum, uint16_t bufsSiz_MB
			   , uint8_t  grpsNum, uint8_t  grpsSiz_pages )
{
    size_t sz;
    sz = 4096*5 + (size_t)bufsNum*(bufsSiz_MB*0x100000) + grpsNum*(grpsSiz_pages*4096);
    return (sz);
}

// The "bufs_largest_multiple" method (using 0xffffffff) allows "easy" "add 1"
// I must do the substract (ie. add negative) by hand.
uint32_t ShmRdWr::bufcnt_add( uint32_t bufcnt, int add )
{   uint32_t tt;
    if (add < 0)
    {   add = -add;
	if ((uint32_t)add > bufcnt)
	{   add -= bufcnt;
	    tt=shm_hdr_ptr_->bufs_largest_multiple - add;
	}
	else
	    tt=bufcnt-add;
    }
    else
	tt=bufcnt+add;
    return (tt%shm_hdr_ptr_->bufs_largest_multiple);
}

int ShmRdWr::bufcnt2idx( uint32_t bufcnt )
{   return (bufcnt % shm_hdr_ptr_->bufs_num);
}

uint32_t ShmRdWr::bufcnt_delta( uint32_t wr, uint32_t rd )
{   return ((wr>=rd)
	    ? wr-rd
	    : wr-rd-shm_hdr_ptr_->largest_zero_offset);
}

timeval& ShmRdWr::set_expire_time( timeval& expire_time, time_t tmo_ms )
{   if (tmo_ms == -1)
    {   expire_time.tv_sec = expire_time.tv_usec = 0;
    }
    else
    {   gettimeofday( &expire_time, NULL );
	time_t seconds = tmo_ms/1000;
	expire_time.tv_sec += seconds;
	time_t ms = tmo_ms - seconds*1000;
	expire_time.tv_usec += ms*1000;
	if (expire_time.tv_usec >= 1000000)
	{   expire_time.tv_usec -= 1000000;
	    ++expire_time.tv_sec;
	}
    }
    DPRINT(("set_expire_time: expire_time.tv_usec=%ld\n", expire_time.tv_usec));
    return (expire_time);
}   // set_expire_time

bool ShmRdWr::is_tmo( timeval& expire_time )
{
    timeval time_now;
    gettimeofday( &time_now, NULL );
    if (  (  (expire_time.tv_sec==time_now.tv_sec)
	   &&(expire_time.tv_usec<=time_now.tv_usec))
	||(expire_time.tv_sec < time_now.tv_sec))
    {   DPRINT(("now:%ld.%06ld  exp:%ld.%06ld\n"
		,time_now.tv_sec,time_now.tv_usec,expire_time.tv_sec,expire_time.tv_usec));
	return (true);
    }
    else
	return (false);
}   // is_tmo

void  ShmRdWr::expire_tv2ts( const timeval& texp, timespec& tdiff )
{   timeval tnow;
    gettimeofday( &tnow, NULL );
    // get diff texp-tnow
    tdiff.tv_nsec  = (texp.tv_usec + 1000000 - tnow.tv_usec)*1000;
    tdiff.tv_sec  = texp.tv_sec - tnow.tv_sec;
    if (tdiff.tv_nsec >= 1000000000)
	tdiff.tv_nsec -= 1000000000;
    else tdiff.tv_sec -= 1;
    if (tdiff.tv_sec < 0)
	tdiff.tv_sec = tdiff.tv_nsec = 0;
    DPRINT(("expire_tv2ts: tdiff.tv_sec=%ld tdiff.tv_nsec=%ld\n"
	    ,tdiff.tv_sec, tdiff.tv_nsec));
    return;
}

// return true if got sem, false otherwise
bool ShmRdWr::semtake( shmrw_gid_t gIdx, timespec& tmo_ts )
{
    sembuf sops;
    sops.sem_num = gIdx;
    sops.sem_op  = -1;
    sops.sem_flg = SEM_UNDO;
    int semsts=0;
#ifdef DARWINBUILD
    semsts=semop( shm_hdr_ptr_->semSetId, &sops, 1 );
#else
    semsts=semtimedop( shm_hdr_ptr_->semSetId, &sops, 1, &tmo_ts );
#endif
    if ((semsts==-1) && (errno==EAGAIN))
    {   DPRINT(("SEMOP TIMEOUT\n"));
	return (false);
    }
    else if (semsts==-1)
    {   perror("semtimedop");
	printf("SEMOP OTHER ERROR=%d semSetId=%d sem_num=%u tmo.sec=%ld tmp.nsec=%ld\n"
		,errno,shm_hdr_ptr_->semSetId,(unsigned short)gIdx,tmo_ts.tv_sec
		,tmo_ts.tv_nsec);
	return (false);
    }
    return true;
}   // semtake

void ShmRdWr::semgive( shmrw_gid_t gIdx )
{   sembuf sops;
    sops.sem_num = gIdx;
    sops.sem_op  = 1;
    sops.sem_flg = SEM_UNDO;
    semop( shm_hdr_ptr_->semSetId, &sops, 1 );
}   // semgive

void ShmRdWr::tprint( const char* fmt, ... )
{	timeval		tv;
        va_list         ap;
    va_start( ap, fmt);
    gettimeofday( &tv, NULL );
    printf( "%ld.%06ld ", tv.tv_sec, tv.tv_usec);
    vprintf( fmt, ap );
    va_end( ap );
}
